from apio.commands.time import cli as cmd_time


def test_time(clirunner, configenv):
    with clirunner.isolated_filesystem():
        configenv()
        result = clirunner.invoke(cmd_time)
        assert result.exit_code != 0
        assert 'Info: No apio.ini file' in result.output
        assert 'Error: insufficient arguments: missing board' in result.output


def test_time_board(clirunner, configenv):
    with clirunner.isolated_filesystem():
        configenv()
        result = clirunner.invoke(cmd_time, ['--board', 'icezum'])
        assert result.exit_code != 0
        if result.exit_code == 1:
            assert 'apio install oss-cad-suite' in result.output



from apio.commands.system import cli as cmd_system


def test_system(clirunner, validate_cliresult, configenv):
    result = clirunner.invoke(cmd_system)
    validate_cliresult(result)


def test_system_lsftdi(clirunner, configenv):
    with clirunner.isolated_filesystem():
        configenv()
        result = clirunner.invoke(cmd_system, ['--lsftdi'])
        assert result.exit_code == 1
        assert 'apio install oss-cad-suite' in result.output


def test_system_lsusb(clirunner, configenv):
    with clirunner.isolated_filesystem():
        configenv()
        result = clirunner.invoke(cmd_system, ['--lsusb'])
        assert result.exit_code == 1
        assert 'apio install oss-cad-suite' in result.output


def test_system_lsserial(clirunner, configenv):
    with clirunner.isolated_filesystem():
        configenv()
        clirunner.invoke(cmd_system, ['--lsserial'])


def test_system_info(clirunner, configenv):
    with clirunner.isolated_filesystem():
        configenv()
        result = clirunner.invoke(cmd_system, ['--info'])
        assert result.exit_code == 0



"""DOC: TODO"""
# -*- coding: utf-8 -*-
# -- This file is part of the Apio project
# -- (C) 2016-2019 FPGAwars
# -- Author Jesús Arroyo
# -- Licence GPLv2

import click

from apio.managers.installer import Installer
from apio.resources import Resources

platforms = [
    "linux",
    "linux_x86_64",
    "linux_i686",
    "linux_armv7l",
    "linux_aarch64",
    "windows",
    "windows_x86",
    "windows_amd64",
    "darwin",
]


@click.command("install")
@click.pass_context
@click.argument("packages", nargs=-1)
@click.option("-a", "--all", is_flag=True, help="Install all packages.")
@click.option(
    "-l", "--list", is_flag=True, help="List all available packages."
)
@click.option(
    "-f", "--force", is_flag=True, help="Force the packages installation."
)
@click.option(
    "-p",
    "--platform",
    type=click.Choice(platforms),
    metavar="",
    help="Set the platform [{}] (Advanced).".format(", ".join(platforms)),
)
def cli(ctx, packages, all, list, force, platform):
    """Install packages.
    Input parameters:
      - packages: List with the names of the packages to install
      - all: Flag: Install all the packages available for that platform
      - list: Flag. List all the packages (installed or not)
      - force: Flag. Force installation
      - platform: Flag. Select platform (advaced. For developers)
    """

    # -- Install the given packages
    if packages:
        for package in packages:

            # -- The instalation is performed by the Installer object
            inst = Installer(package, platform, force)

            inst.install()

    # -- Install all the available packages
    elif all:  # pragma: no cover

        # -- Get all the resources
        resources = Resources(platform)

        # -- Get all the packages
        packages = resources.packages

        # -- Install all the packages... excepto the obolete one
        for package in packages:

            # -- do NOT install the obsolete packages
            if package not in resources.obsolete_pkgs:
                Installer(package, platform, force).install()

    # -- List all the packages (installed or not)
    elif list:
        # -- Get all the resources
        resources = Resources(platform)

        # -- List the packages
        resources.list_packages()

    # -- Invalid option. Just show the help
    else:
        click.secho(ctx.get_help())



"""DOC:TODO"""
# -*- coding: utf-8 -*-
# -- This file is part of the Apio project
# -- (C) 2016-2019 FPGAwars
# -- Author Jesús Arroyo
# -- Licence GPLv2

# --------------------------------------------
# -- Apio ENTRY POINT!!!
# --------------------------------------------


from os import listdir
from os.path import isfile
from sys import exit as sys_exit
import click

from apio import util


# -- The commands are python modules located in the commands folder
# -- Get the absolute path to all the commands
commands_folder = util.get_folder("commands")


# -- Auxiliary function used sfor reformating the Help string output
def find_commands_help(help_str, commands):
    """Extract the commands from the help string
    Return a list with the command + description
    """
    commands_help = []
    for line in list(help_str):
        for command in commands:
            if " " + command in line:
                if line in help_str:
                    help_str.remove(line)
                    commands_help.append(line)
    return commands_help


# -- Main Click Class
# -- It is extended for including methods for getting and listing the commands
class ApioCLI(click.MultiCommand):
    """DOC:TODO"""

    # -- Return  a list of all the available commands
    def list_commands(self, ctx):
        cmd_list = []

        # -- The supported command are python files located
        # -- in the commands folder
        # -- Except __init__.py
        for filename in listdir(commands_folder):
            if filename.startswith("__init__"):
                continue
            if filename.endswith(".py"):
                cmd_list.append(filename[:-3])

        cmd_list.sort()
        return cmd_list

    # -- Return the code function (cli) of the command name
    # -- This cli function is called whenever the name command
    # -- is issued
    # pylint: disable=arguments-differ
    def get_command(self, ctx, cmd_name):
        nnss = {}

        # -- Get the full filename of the python file for the command name
        filename = util.safe_join(commands_folder, cmd_name + ".py")

        # -- Return the cli function of the command
        if isfile(filename):
            with open(filename) as file:
                code = compile(file.read(), filename, "exec")
                eval(code, nnss, nnss)

        return nnss.get("cli")


@click.command(cls=ApioCLI, invoke_without_command=True)
@click.pass_context
@click.version_option()
def cli(ctx):

    # -- No command typed: show help
    if ctx.invoked_subcommand is None:

        # -- The help string is automatically generated by Click
        # -- It could be directly printed on the console but...
        _help = ctx.get_help()

        # -- Let's change the format so that the commands are divided
        # -- into three categories: Project, Setup and Utilitys
        _help = _help.split("\n")

        # -- Setup commands
        setup_help = find_commands_help(
            _help, ["drivers", "init", "install", "uninstall"]
        )

        # -- Utility commands
        util_help = find_commands_help(
            _help, ["boards", "config", "examples", "raw", "system", "upgrade"]
        )

        # -- Reformat the Help string
        _help = "\n".join(_help)
        _help = _help.replace("Commands:\n", "Project commands:\n")
        _help += "\n\nSetup commands:\n"
        _help += "\n".join(setup_help)
        _help += "\n\nUtility commands:\n"
        _help += "\n".join(util_help)
        _help += "\n"

        click.secho(_help)

    # -- If there is a command, it is executed when this function is finished
    # -- Debug: print the command invoked
    # print(f"{ctx.invoked_subcommand}")


# -- This is NOT used
if __name__ == "__main__":  # pragma: no cover
    sys_exit()



from apio.commands.install import cli as cmd_install


def test_install(clirunner, validate_cliresult):
    result = clirunner.invoke(cmd_install)
    validate_cliresult(result)


def test_install_list(clirunner, validate_cliresult, configenv):
    with clirunner.isolated_filesystem():
        configenv()
        result = clirunner.invoke(cmd_install, ['--list'])
        validate_cliresult(result)


def test_install_wrong_package(clirunner, configenv):
    with clirunner.isolated_filesystem():
        configenv()
        result = clirunner.invoke(cmd_install, ['missing_package'])
        assert result.exit_code == 1
        assert 'Error: no such package' in result.output



# -*- coding: utf-8 -*-
# -- This file is part of the Apio project
# -- (C) 2016-2019 FPGAwars
# -- Author Jesús Arroyo
# -- Licence GPLv2
# -- Derived from:
# ---- Platformio project
# ---- (C) 2014-2016 Ivan Kravets <me@ikravets.com>
# ---- Licence Apache v2


from email.utils import parsedate_tz
from math import ceil
from time import mktime
import requests
import click


from apio import util

requests.packages.urllib3.disable_warnings()


class FDUnrecognizedStatusCode(util.ApioException):

    MESSAGE = "Got an unrecognized status code '{0}' when downloaded {1}"


class FileDownloader:

    CHUNK_SIZE = 1024

    def __init__(self, url, dest_dir=None):
        self._url = url
        self._fname = url.split("/")[-1]

        self._destination = self._fname
        if dest_dir:
            self.set_destination(util.safe_join(dest_dir, self._fname))

        self._progressbar = None
        self._request = None

        # make connection
        self._request = requests.get(url, stream=True)
        if self._request.status_code != 200:
            raise FDUnrecognizedStatusCode(self._request.status_code, url)

    def set_destination(self, destination):
        self._destination = destination

    def get_filepath(self):
        return self._destination

    def get_lmtime(self):
        if "last-modified" in self._request.headers:
            return self._request.headers.get("last-modified")
        return None

    def get_size(self):
        return int(self._request.headers.get("content-length"))

    def start(self):
        itercontent = self._request.iter_content(chunk_size=self.CHUNK_SIZE)
        with open(self._destination, "wb") as file:
            chunks = int(ceil(self.get_size() / float(self.CHUNK_SIZE)))

            with click.progressbar(length=chunks, label="Downloading") as pbar:
                for _ in pbar:
                    file.write(next(itercontent))
        self._request.close()

        self._preserve_filemtime(self.get_lmtime())

    def _preserve_filemtime(self, lmdate):
        if lmdate is not None:
            timedata = parsedate_tz(lmdate)
            lmtime = mktime(timedata[:9])
            util.change_filemtime(self._destination, lmtime)

    def __del__(self):
        if self._request:
            self._request.close()



from apio.commands.upload import cli as cmd_upload


def test_upload(clirunner, configenv):
    with clirunner.isolated_filesystem():
        configenv()
        result = clirunner.invoke(cmd_upload)
        assert result.exit_code == 1
        assert 'Info: No apio.ini file' in result.output
        assert 'Error: insufficient arguments: missing board' in result.output


def test_upload_serial_port(clirunner, configenv):
    with clirunner.isolated_filesystem():
        configenv()
        result = clirunner.invoke(cmd_upload, ['--serial-port', 'COM0'])
        assert result.exit_code == 1
        assert 'Info: No apio.ini file' in result.output
        assert 'Error: insufficient arguments: missing board' in result.output


def test_upload_ftdi_id(clirunner, configenv):
    with clirunner.isolated_filesystem():
        configenv()
        result = clirunner.invoke(cmd_upload, ['--ftdi-id', '0'])
        assert result.exit_code == 1
        assert 'Info: No apio.ini file' in result.output
        assert 'Error: insufficient arguments: missing board' in result.output


def test_upload_sram(clirunner, configenv):
    with clirunner.isolated_filesystem():
        configenv()
        result = clirunner.invoke(cmd_upload, ['--sram'])
        assert result.exit_code == 1
        assert 'Info: No apio.ini file' in result.output
        assert 'Error: insufficient arguments: missing board' in result.output


def test_upload_board(clirunner, configenv):
    with clirunner.isolated_filesystem():
        configenv()
        result = clirunner.invoke(cmd_upload, ['--board', 'icezum'])
        assert result.exit_code == 1


def test_upload_board_serial_port(clirunner, configenv):
    with clirunner.isolated_filesystem():
        configenv()
        result = clirunner.invoke(cmd_upload, [
            '--board', 'icezum', '--serial-port', 'COM0'])
        assert result.exit_code == 1


def test_upload_board_ftdi_id(clirunner, configenv):
    with clirunner.isolated_filesystem():
        configenv()
        result = clirunner.invoke(cmd_upload, [
            '--board', 'icezum', '--ftdi-id', '0'])
        assert result.exit_code == 1


def test_upload_board_sram(clirunner, configenv):
    with clirunner.isolated_filesystem():
        configenv()
        result = clirunner.invoke(cmd_upload, [
            '--board', 'icezum', '--sram'])
        assert result.exit_code == 1



import pytest
from apio.commands.upgrade import cli as cmd_upgrade


def test_upgrade(clirunner, validate_cliresult, offline):
    if offline:
        pytest.skip('requires internet connection')

    result = clirunner.invoke(cmd_upgrade)
    validate_cliresult(result)



# -*- coding: utf-8 -*-
# -- This file is part of the Apio project
# -- (C) 2016-2019 FPGAwars
# -- Author Jesús Arroyo
# -- Licence GPLv2

import sys

import click

from apio.managers.scons import SCons
from apio.managers.drivers import Drivers

# Python3 compat
if sys.version_info > (3, 0):
    unicode = str


@click.command("upload")
@click.pass_context
@click.option(
    "-b", "--board", type=unicode, metavar="board", help="Set the board."
)
@click.option(
    "--serial-port",
    type=unicode,
    metavar="serial-port",
    help="Set the serial port.",
)
@click.option(
    "--ftdi-id", type=unicode, metavar="ftdi-id", help="Set the FTDI id."
)
@click.option("-s", "--sram", is_flag=True, help="Perform SRAM programming.")
@click.option("-f", "--flash", is_flag=True, help="Perform FLASH programming.")
@click.option(
    "-p",
    "--project-dir",
    type=unicode,
    metavar="path",
    help="Set the target directory for the project.",
)
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    help="Show the entire output of the command.",
)
@click.option(
    "--verbose-yosys",
    is_flag=True,
    help="Show the yosys output of the command.",
)
@click.option(
    "--verbose-pnr", is_flag=True, help="Show the pnr output of the command."
)
def cli(
    ctx,
    board,
    serial_port,
    ftdi_id,
    sram,
    flash,
    project_dir,
    verbose,
    verbose_yosys,
    verbose_pnr,
):
    """Upload the bitstream to the FPGA."""

    drivers = Drivers()
    drivers.pre_upload()
    # Run scons

    exit_code = SCons(project_dir).upload(
        {
            "board": board,
            "verbose": {
                "all": verbose,
                "yosys": verbose_yosys,
                "pnr": verbose_pnr,
            },
        },
        serial_port,
        ftdi_id,
        sram,
        flash,
    )
    drivers.post_upload()
    ctx.exit(exit_code)


# Advanced notes: https://github.com/FPGAwars/apio/wiki/Commands#apio-upload



from os import getcwd, mkdir
from os.path import join, isfile, getsize

from apio.commands.init import cli as cmd_init


def validate_apio_ini(current_dir):
    path = join(current_dir, 'apio.ini')
    assert isfile(path) and getsize(path) > 0


def validate_scons(apioproject_dir):
    path = join(apioproject_dir, 'SConstruct')
    assert isfile(path) and getsize(path) > 0


def test_init(clirunner, validate_cliresult):
    result = clirunner.invoke(cmd_init)
    validate_cliresult(result)


def test_init_wrong_board(clirunner, validate_cliresult):
    with clirunner.isolated_filesystem():
        result = clirunner.invoke(cmd_init, ['--board', 'missed_board'])
        assert result.exit_code == 1
        assert 'Error: no such board' in result.output


def test_init_board(clirunner, validate_cliresult):
    with clirunner.isolated_filesystem():

        # apio init --board icezum
        result = clirunner.invoke(cmd_init, ['--board', 'icezum'])
        validate_cliresult(result)
        validate_apio_ini(getcwd())
        assert 'Creating apio.ini file ...' in result.output
        assert 'has been successfully created!' in result.output

        # apio init --board icezum
        result = clirunner.invoke(cmd_init, ['--board', 'icezum'], input='y')
        validate_cliresult(result)
        validate_apio_ini(getcwd())
        assert 'Warning' in result.output
        assert 'file already exists' in result.output
        assert 'Do you want to replace it?' in result.output
        assert 'has been successfully created!' in result.output


def test_init_scons(clirunner, validate_cliresult):
    with clirunner.isolated_filesystem():

        # apio init --scons
        result = clirunner.invoke(cmd_init, ['--scons'])
        validate_cliresult(result)
        validate_scons(getcwd())
        assert 'Creating SConstruct file ...' in result.output
        assert 'has been successfully created!' in result.output

        # apio init --scons
        result = clirunner.invoke(cmd_init, ['--scons'], input='y')
        validate_cliresult(result)
        validate_scons(getcwd())
        assert 'Warning' in result.output
        assert 'file already exists' in result.output
        assert 'Do you want to replace it?' in result.output
        assert 'has been successfully created!' in result.output


def test_init_scons_project_dir(clirunner, validate_cliresult):
    with clirunner.isolated_filesystem():
        dir_name = 'tmp'
        mkdir(dir_name)

        # apio init --scons --project-dir=tmp
        result = clirunner.invoke(cmd_init, ['--scons', '--project-dir=tmp'])
        validate_cliresult(result)
        validate_scons(join(getcwd(), dir_name))
        assert 'Creating SConstruct file ...' in result.output
        assert 'has been successfully created!' in result.output



# -*- coding: utf-8 -*-
# -- This file is part of the Apio project
# -- (C) 2016-2019 FPGAwars
# -- Author Jesús Arroyo
# -- Licence GPLv2

import click

from apio.util import get_systype
from apio.managers.system import System


@click.command("system")
@click.pass_context
@click.option(
    "--lsftdi", is_flag=True, help="List all connected FTDI devices."
)
@click.option("--lsusb", is_flag=True, help="List all connected USB devices.")
@click.option(
    "--lsserial", is_flag=True, help="List all connected Serial devices."
)
@click.option("-i", "--info", is_flag=True, help="Show system information.")
def cli(ctx, lsftdi, lsusb, lsserial, info):
    """System tools.\n
    Install with `apio install system`"""

    exit_code = 0

    if lsftdi:
        exit_code = System().lsftdi()
    elif lsusb:
        exit_code = System().lsusb()
    elif lsserial:
        exit_code = System().lsserial()
    elif info:
        click.secho("Platform: ", nl=False)
        click.secho(get_systype(), fg="yellow")
    else:
        click.secho(ctx.get_help())

    ctx.exit(exit_code)



# -*- coding: utf-8 -*-
# -- This file is part of the Apio project
# -- (C) 2016-2019 FPGAwars
# -- Author Jesús Arroyo
# -- Licence GPLv2

import sys
import click

from apio.managers.examples import Examples


# Python3 compat
if sys.version_info > (3, 0):
    unicode = str


@click.command("examples")
@click.pass_context
@click.option(
    "-l", "--list", is_flag=True, help="List all available examples."
)
@click.option(
    "-d",
    "--dir",
    type=unicode,
    metavar="name",
    help="Copy the selected example directory.",
)
@click.option(
    "-f",
    "--files",
    type=unicode,
    metavar="name",
    help="Copy the selected example files.",
)
@click.option(
    "-p",
    "--project-dir",
    type=unicode,
    metavar="path",
    help="Set the target directory for the examples.",
)
@click.option(
    "-n",
    "--sayno",
    is_flag=True,
    help="Automatically answer NO to all the questions.",
)
def cli(ctx, list, dir, files, project_dir, sayno):
    """Manage verilog examples.\n
    Install with `apio install examples`"""

    exit_code = 0

    if list:
        exit_code = Examples().list_examples()
    elif dir:
        exit_code = Examples().copy_example_dir(dir, project_dir, sayno)
    elif files:
        exit_code = Examples().copy_example_files(files, project_dir, sayno)
    else:
        click.secho(ctx.get_help())
        click.secho(Examples().examples_of_use_cad())

    ctx.exit(exit_code)



# -*- coding: utf-8 -*-
# -- This file is part of the Apio project
# -- (C) 2016-2019 FPGAwars
# -- Author Jesús Arroyo
# -- Licence GPLv2

import sys
import click

from apio.managers.project import Project

# Python3 compat
if sys.version_info > (3, 0):
    unicode = str


@click.command("init")
@click.pass_context
@click.option(
    "-s", "--scons", is_flag=True, help="Create default SConstruct file."
)
@click.option(
    "-b",
    "--board",
    type=unicode,
    metavar="board",
    help="Create init file with the selected board.",
)
@click.option(
    "-p",
    "--project-dir",
    type=unicode,
    metavar="path",
    help="Set the target directory for the project.",
)
@click.option(
    "-y",
    "--sayyes",
    is_flag=True,
    help="Automatically answer YES to all the questions.",
)
def cli(ctx, board, scons, project_dir, sayyes):
    """Manage apio projects."""

    if scons:
        Project().create_sconstruct(project_dir, "ice40", sayyes)
    elif board:
        Project().create_ini(board, project_dir, sayyes)
    else:
        click.secho(ctx.get_help())



from apio.commands.examples import cli as cmd_examples


def test_examples(clirunner, validate_cliresult, configenv):
    result = clirunner.invoke(cmd_examples)
    validate_cliresult(result)


def test_examples_list(clirunner, configenv):
    with clirunner.isolated_filesystem():
        configenv()
        result = clirunner.invoke(cmd_examples, ['--list'])
        assert result.exit_code == 1
        assert 'apio install examples' in result.output


def test_examples_dir(clirunner, configenv):
    with clirunner.isolated_filesystem():
        configenv()
        result = clirunner.invoke(cmd_examples, ['--dir', 'dir'])
        assert result.exit_code == 1
        assert 'apio install examples' in result.output


def test_examples_files(clirunner, configenv):
    with clirunner.isolated_filesystem():
        configenv()
        result = clirunner.invoke(cmd_examples, ['--files', 'file'])
        assert result.exit_code == 1
        assert 'apio install examples' in result.output



"""Resources module"""
# -*- coding: utf-8 -*-
# -- This file is part of the Apio project
# -- (C) 2016-2019 FPGAwars
# -- Author Jesús Arroyo
# -- Licence GPLv2

# ---------- RESOURCES

# ---------------------------------------
# ---- File: resources/packages.json
# --------------------------------------
# -- This file contains all the information regarding the available apio
# -- packages: Repository, version, name...
# -- This information is access through the Resources.packages method

# -----------------------------------------
# ---- File: resources/boads.json
# -----------------------------------------
# -- Information about all the supported board
# -- names, fpga family, programmer, ftdi description, vendor id, product id

import json
from collections import OrderedDict
import shutil
import click

from apio import util
from apio.profile import Profile

# -- Info message
BOARDS_MSG = (
    """
Use `apio init --board <boardname>` to create a new apio """
    """project for that board"""
)

# -- Packages marked as obsoletes
# -- The value is the replacement package (if any)
OBSOLETE_PKGS = {
    "system": "oss-cad-suite",
    "yosys": "oss-cad-suite",
    "ice40": "oss-cad-suite",
    "scons": "",
}


class Resources:
    """Resource manager. Class for accesing to all the resources"""

    def __init__(self, platform=""):

        # -- Read the packages information
        self.packages = self._load_resource("packages")

        # -- Read the boards information
        self.boards = self._load_resource("boards")

        self.fpgas = self._load_resource("fpgas")
        self.programmers = self._load_resource("programmers")
        self.distribution = self._load_resource("distribution")

        # Check available packages
        self.packages = self._check_packages(self.packages, platform)

        # Sort resources
        self.packages = OrderedDict(
            sorted(self.packages.items(), key=lambda t: t[0])
        )

        # -- Obsolete packages
        self.obsolete_pkgs = OBSOLETE_PKGS

        self.boards = OrderedDict(
            sorted(self.boards.items(), key=lambda t: t[0])
        )
        self.fpgas = OrderedDict(
            sorted(self.fpgas.items(), key=lambda t: t[0])
        )
        self.profile = None

    @staticmethod
    def _load_resource(name):
        """Load the resources from a given json file
        * Name: Name of the file without extension:
         * boards: Load the boards
         * distribution
         * fpgas
         * packages
         * programmers
        """

        # -- Default return value: no resource
        resource = None

        # -- Build the filepath: Ex. resources/fpgas.json
        filepath = util.safe_join(util.get_folder("resources"), name + ".json")

        # -- Open the json file and convert it to an object
        with open(filepath, "r") as file:
            resource = json.loads(file.read())

        # -- Return the object for the resource
        return resource

    def get_package_release_name(self, package):
        """DOC: TODO"""

        return self.packages.get(package).get("release").get("package_name")

    def get_packages(self):
        """DOC: TODO"""

        # Classify packages
        installed_packages = []
        notinstalled_packages = []

        for package in self.packages:
            data = {
                "name": package,
                "version": None,
                "description": self.packages.get(package).get("description"),
            }
            if package in self.profile.packages:
                data["version"] = self.profile.get_package_version(
                    package, self.get_package_release_name(package)
                )
                installed_packages += [data]
            else:
                notinstalled_packages += [data]

        for package in self.profile.packages:
            if package not in self.packages:
                data = {
                    "name": package,
                    "version": "Unknown",
                    "description": "Unknown deprecated package",
                }
                installed_packages += [data]

        return installed_packages, notinstalled_packages

    def list_packages(self, installed=True, notinstalled=True):
        """Return a list with all the installed/notinstalled packages"""

        self.profile = Profile()

        # Classify packages
        installed_packages, notinstalled_packages = self.get_packages()

        # Print tables
        terminal_width, _ = shutil.get_terminal_size()

        if installed and installed_packages:

            # - Print installed packages table
            click.echo("\nInstalled packages:\n")

            package_list_tpl = "{name:20} {description:30} {version:<8}"

            click.echo("-" * terminal_width)
            click.echo(
                package_list_tpl.format(
                    name=click.style("Name", fg="cyan"),
                    version="Version",
                    description="Description",
                )
            )
            click.echo("-" * terminal_width)

            for package in installed_packages:
                click.echo(
                    package_list_tpl.format(
                        name=click.style(package.get("name"), fg="cyan"),
                        version=package.get("version"),
                        description=package.get("description"),
                    )
                )

        if notinstalled and notinstalled_packages:

            # - Print not installed packages table
            click.echo("\nNot installed packages:\n")

            package_list_tpl = "{name:20} {description:30}"

            click.echo("-" * terminal_width)
            click.echo(
                package_list_tpl.format(
                    name=click.style("Name", fg="yellow"),
                    description="Description",
                )
            )
            click.echo("-" * terminal_width)

            for package in notinstalled_packages:
                click.echo(
                    package_list_tpl.format(
                        name=click.style(package.get("name"), fg="yellow"),
                        description=package.get("description"),
                    )
                )

        # -- Print the obsolete packages
        click.echo("\nObsolete packages:\n")
        package_list_tpl = "{name:20} {description:30}"
        click.echo("-" * terminal_width)
        click.echo(
            package_list_tpl.format(
                name=click.style("Old package", fg="yellow"),
                description="New package",
            )
        )
        click.echo("-" * terminal_width)

        for package, new_package in OBSOLETE_PKGS.items():
            click.echo(
                package_list_tpl.format(
                    name=click.style(package, fg="yellow"),
                    description=new_package,
                )
            )

        click.echo("\n")

    def list_boards(self):
        """Return a list with all the supported boards"""

        # Print table
        click.echo("\nSupported boards:\n")

        board_list_tpl = (
            "{board:25} {fpga:30} {arch:<8} "
            + "{type:<12} {size:<5} {pack:<10}"
        )
        terminal_width, _ = shutil.get_terminal_size()

        click.echo("-" * terminal_width)
        click.echo(
            board_list_tpl.format(
                board=click.style("Board", fg="cyan"),
                fpga="FPGA",
                arch="Arch",
                type="Type",
                size="Size",
                pack="Pack",
            )
        )
        click.echo("-" * terminal_width)

        for board in self.boards:
            fpga = self.boards.get(board).get("fpga")
            click.echo(
                board_list_tpl.format(
                    board=click.style(board, fg="cyan"),
                    fpga=fpga,
                    arch=self.fpgas.get(fpga).get("arch"),
                    type=self.fpgas.get(fpga).get("type"),
                    size=self.fpgas.get(fpga).get("size"),
                    pack=self.fpgas.get(fpga).get("pack"),
                )
            )

        click.secho(BOARDS_MSG, fg="green")

    def list_fpgas(self):
        """Return a list with all the supported FPGAs"""

        # Print table
        click.echo("\nSupported FPGAs:\n")

        fpga_list_tpl = "{fpga:40} {arch:<8} {type:<12} {size:<5} {pack:<10}"
        terminal_width, _ = shutil.get_terminal_size()

        click.echo("-" * terminal_width)
        click.echo(
            fpga_list_tpl.format(
                fpga=click.style("FPGA", fg="cyan"),
                type="Type",
                arch="Arch",
                size="Size",
                pack="Pack",
            )
        )
        click.echo("-" * terminal_width)

        for fpga in self.fpgas:
            click.echo(
                fpga_list_tpl.format(
                    fpga=click.style(fpga, fg="cyan"),
                    arch=self.fpgas.get(fpga).get("arch"),
                    type=self.fpgas.get(fpga).get("type"),
                    size=self.fpgas.get(fpga).get("size"),
                    pack=self.fpgas.get(fpga).get("pack"),
                )
            )

    @staticmethod
    def _check_packages(packages, current_platform=""):
        filtered_packages = {}
        for pkg in packages.keys():
            check = True
            release = packages.get(pkg).get("release")
            if "available_platforms" in release:
                platforms = release.get("available_platforms")
                check = False
                current_platform = current_platform or util.get_systype()
                for platform in platforms:
                    check |= current_platform in platform
            if check:
                filtered_packages[pkg] = packages.get(pkg)
        return filtered_packages



import pytest

from apio.api import api_request


def test_api_request(capsys, offline):
    if offline:
        pytest.skip('requires internet connection')

    with pytest.raises(SystemExit):
        api_request('missing_command')
    out, err = capsys.readouterr()
    assert "Error: " in out



"""Implementation of the Apio system commands"""
# -*- coding: utf-8 -*-
# -- This file is part of the Apio project
# -- (C) 2016-2019 FPGAwars
# -- Author Jesús Arroyo
# -- Licence GPLv2

import re
import platform
from pathlib import Path
from os.path import isfile
import click

from apio import util
from apio.profile import Profile
from apio.resources import Resources


class System:  # pragma: no cover
    """System class. Managing and execution of the system commands"""

    def __init__(self):

        # -- Read the profile from the file
        profile = Profile()

        # -- Read the resources from the corresponding files
        resources = Resources()

        # -- This command is called system
        self.name = "system"

        # -- Package name: apio package were all the system commands
        # -- are located
        # -- From apio > 0.7 the system tools are located inside the
        # -- oss-cad-suite
        self.package_name = "oss-cad-suite"

        # -- Get the installed package versions
        self.version = util.get_package_version(self.name, profile)

        # -- Get the spected versions
        self.spec_version = util.get_package_spec_version(self.name, resources)

        # -- Windows: Executables should end with .exe
        self.ext = ""
        if platform.system() == "Windows":
            self.ext = ".exe"

    def lsusb(self):
        """Run the lsusb system command"""

        returncode = 1
        result = self._run_command("lsusb")

        if result:
            returncode = result.get("returncode")

        return returncode

    def lsftdi(self):
        """DOC: TODO"""

        returncode = 1
        result = self._run_command("lsftdi")

        if result:
            returncode = result.get("returncode")

        return returncode

    @staticmethod
    def lsserial():
        """DOC: TODO"""

        returncode = 0
        serial_ports = util.get_serial_ports()
        click.secho(
            "Number of Serial devices found: {}\n".format(len(serial_ports))
        )

        for serial_port in serial_ports:
            port = serial_port.get("port")
            description = serial_port.get("description")
            hwid = serial_port.get("hwid")
            click.secho(port, fg="cyan")
            click.secho("Description: {}".format(description))
            click.secho("Hardware info: {}\n".format(hwid))

        return returncode

    def get_usb_devices(self):
        """DOC: TODO"""

        usb_devices = []
        result = self._run_command("lsusb", silent=True)

        if result and result.get("returncode") == 0:
            usb_devices = self._parse_usb_devices(result.get("out"))
        else:
            raise Exception

        return usb_devices

    def get_ftdi_devices(self):
        """DOC: TODO"""

        ftdi_devices = []
        result = self._run_command("lsftdi", silent=True)

        if result and result.get("returncode") == 0:
            ftdi_devices = self._parse_ftdi_devices(result.get("out"))
        else:
            raise Exception

        return ftdi_devices

    def _run_command(self, command, silent=False):
        result = {}

        print(f"(DEBUG) Run Command: {command}")

        # From apio >= 0.7.0, the system tools are locate in the
        # oss-cad-suite package instead of the system package
        # So first let's try to execute them from there

        # -- Get the package base dir
        system_base_dir = util.get_package_dir("tools-oss-cad-suite")
        print(f"(DEBUG) System_base_dir: {system_base_dir}")

        # -- Get the folder were the binary file is locateds
        system_bin_dir = Path(system_base_dir) / "bin"
        print(f"(DEBUG) System bin dir: {system_bin_dir}")

        # -- Get the executable filename
        executable_file = system_bin_dir / (command + self.ext)
        print(f"(DEBUG) Executable file: {executable_file}")

        # -- Set the stdout and stderr for executing the command
        on_stdout = None if silent else self._on_stdout
        on_stderr = self._on_stderr

        # -- Check if the executable exists
        if isfile(executable_file):

            # -- Execute the command!
            result = util.exec_command(
                executable_file,
                stdout=util.AsyncPipe(on_stdout),
                stderr=util.AsyncPipe(on_stderr),
            )

            # -- Return the result of the execution
            return result

        # -- The command does not exist in the tools-oss-cad-suite package
        # -- Try with the tool-system package (old-package,
        # -- (for compatibility reasons)

        # -- Get the package base dir
        system_base_dir = util.get_package_dir("tools-system")
        print(f"System_base_dir: {system_base_dir}")

        # -- Get the folder were the binary file is locateds
        system_bin_dir = Path(system_base_dir) / "bin"
        print(f"System bin dir: {system_bin_dir}")

        # -- Get the executable filename
        executable_file = system_bin_dir / (command + self.ext)
        print(f"Executable file: {executable_file}")

        # -- Check if the executable exists
        if isfile(executable_file):

            # -- Execute the command!
            result = util.exec_command(
                executable_file,
                stdout=util.AsyncPipe(on_stdout),
                stderr=util.AsyncPipe(on_stderr),
            )

            # -- Return the result of the execution
            return result

        # -- The command was not in the oss-cad-suit package
        # -- The command was not in the old system package
        # -- Show the error message and a hint on how to install the package
        print("ERROR!!!!!!")
        util.show_package_path_error(self.package_name)
        util.show_package_install_instructions(self.package_name)

        return None

    @staticmethod
    def _on_stdout(line):
        click.secho(line)

    @staticmethod
    def _on_stderr(line):
        click.secho(line, fg="red")

    @staticmethod
    def _parse_usb_devices(text):
        pattern = r"(?P<hwid>[a-f0-9]{4}:[a-f0-9]{4}?)\s"
        hwids = re.findall(pattern, text)

        usb_devices = []

        for hwid in hwids:
            usb_device = {"hwid": hwid}
            usb_devices.append(usb_device)

        return usb_devices

    @staticmethod
    def _parse_ftdi_devices(text):
        pattern = r"Number\sof\sFTDI\sdevices\sfound:\s(?P<n>\d+?)\n"
        match = re.search(pattern, text)
        num = int(match.group("n")) if match else 0

        pattern = r".*Checking\sdevice:\s(?P<index>.*?)\n.*"
        index = re.findall(pattern, text)

        pattern = r".*Manufacturer:\s(?P<n>.*?),.*"
        manufacturer = re.findall(pattern, text)

        pattern = r".*Description:\s(?P<n>.*?)\n.*"
        description = re.findall(pattern, text)

        ftdi_devices = []

        for i in range(num):
            ftdi_device = {
                "index": index[i],
                "manufacturer": manufacturer[i],
                "description": description[i],
            }
            ftdi_devices.append(ftdi_device)

        return ftdi_devices



from apio.commands.build import cli as cmd_build
from apio.commands.init import cli as cmd_init


def test_build(clirunner, configenv):
    with clirunner.isolated_filesystem():
        configenv()
        result = clirunner.invoke(cmd_build)
        assert result.exit_code != 0
        assert 'Info: No apio.ini file' in result.output
        assert 'Error: insufficient arguments: missing board' in result.output


def test_build_board(clirunner, configenv):
    with clirunner.isolated_filesystem():
        configenv()
        result = clirunner.invoke(cmd_build, ['--board', 'icezum'])
        assert result.exit_code != 0
        if result.exit_code == 1:
            assert 'install oss-cad-suite' in result.output


def test_build_complete(clirunner, configenv):
    with clirunner.isolated_filesystem():
        configenv()

        # apio build --board icestick
        result = clirunner.invoke(cmd_build, ['--board', 'icestick'])
        assert result.exit_code != 0

        # apio build --fpga iCE40-HX1K-VQ100
        result = clirunner.invoke(cmd_build, ['--fpga', 'iCE40-HX1K-VQ100'])
        assert result.exit_code != 0

        # apio build --type lp --size 8k --pack cm225:4k
        result = clirunner.invoke(cmd_build, [
            '--type', 'lp', '--size', '8k', '--pack', 'cm225:4k'])
        assert result.exit_code != 0

        # apio build --board icezum --size 1k
        result = clirunner.invoke(cmd_build, [
            '--board', 'icezum', '--size', '1k'])
        assert result.exit_code != 0
        assert 'Warning: redundant arguments: size' in result.output

        # apio build --board icezum --fpga iCE40-HX1K-TQ144 --type hx
        result = clirunner.invoke(cmd_build, [
            '--board', 'icezum', '--fpga', 'iCE40-HX1K-TQ144', '--type', 'hx'])
        assert result.exit_code != 0
        assert 'Warning: redundant arguments: fpga, type' in result.output

        # apio build --board icezum --pack tq144
        result = clirunner.invoke(cmd_build, [
            '--board', 'icezum', '--pack', 'tq144'])
        assert result.exit_code != 0
        assert 'Warning: redundant arguments: pack' in result.output

        # apio build --fpga iCE40-HX1K-TQ144 --pack tq144 --size 1k
        result = clirunner.invoke(cmd_build, [
            '--fpga', 'iCE40-HX1K-TQ144', '--pack', 'tq144', '--size', '1k'])
        assert result.exit_code != 0
        assert 'Warning: redundant arguments: size, pack' in result.output

        # apio build --fpga iCE40-HX1K-TQ144 --type hx
        result = clirunner.invoke(cmd_build, [
            '--fpga', 'iCE40-HX1K-TQ144', '--type', 'hx'])
        assert result.exit_code != 0
        assert 'Warning: redundant arguments: type' in result.output

        # apio build --board icezum --size 8k
        result = clirunner.invoke(cmd_build, [
            '--board', 'icezum', '--size', '8k'])
        assert result.exit_code != 0
        assert 'Error: contradictory arguments: size' in result.output

        # apio build --board icezum --fpga iCE40-HX1K-TQ144 --type lp
        result = clirunner.invoke(cmd_build, [
            '--board', 'icezum', '--fpga', 'iCE40-HX1K-TQ144', '--type', 'lp'])
        assert result.exit_code != 0
        assert 'Warning: redundant arguments: fpga' in result.output
        assert 'Error: contradictory arguments: type' in result.output

        # apio build --board icezum --fpga iCE40-HX1K-VQ100
        result = clirunner.invoke(cmd_build, [
            '--board', 'icezum', '--fpga', 'iCE40-HX1K-VQ100'])
        assert result.exit_code != 0
        assert 'Error: contradictory arguments: fpga' in result.output

        # apio build --fpga iCE40-HX1K-TQ144 --type lp --size 8k
        result = clirunner.invoke(cmd_build, [
            '--fpga', 'iCE40-HX1K-TQ144', '--type', 'lp', '--size', '8k'])
        assert result.exit_code != 0
        assert 'Error: contradictory arguments: size, type' in result.output

        # apio build --fpga iCE40-HX1K-TQ144 --pack vq100
        result = clirunner.invoke(cmd_build, [
            '--fpga', 'iCE40-HX1K-TQ144', '--pack', 'vq100'])
        assert result.exit_code != 0
        assert 'Error: contradictory arguments: pack' in result.output

        # apio build --board icezum --pack vq100
        result = clirunner.invoke(cmd_build, [
            '--board', 'icezum', '--pack', 'vq100'])
        assert result.exit_code != 0
        assert 'Error: contradictory arguments: pack' in result.output

        # apio build --size 8k
        result = clirunner.invoke(cmd_build, ['--size', '8k'])
        assert result.exit_code != 0
        assert 'Error: insufficient arguments: missing type, pack' \
            in result.output

        # apio build --type lp
        result = clirunner.invoke(cmd_build, ['--type', 'lp'])
        assert result.exit_code != 0
        assert 'Error: insufficient arguments: missing size, pack' \
            in result.output

        # apio build --type lp --size 8k
        result = clirunner.invoke(cmd_build, ['--type', 'lp', '--size', '8k'])
        assert result.exit_code != 0
        assert 'Error: insufficient arguments: missing pack' in result.output

        # apio build --board icefake
        result = clirunner.invoke(cmd_build, ['--board', 'icefake'])
        assert result.exit_code != 0
        assert 'Error: unknown board: icefake' in result.output

        # apio build --board icefake --fpga iCE40-HX1K-TQ144
        result = clirunner.invoke(cmd_build, [
            '--board', 'icefake', '--fpga', 'iCE40-HX1K-TQ144'])
        assert result.exit_code != 0
        assert 'Error: unknown board: icefake' in result.output

        # apio build --fpga iCE40-FAKE
        result = clirunner.invoke(cmd_build, ['--fpga', 'iCE40-FAKE'])
        assert result.exit_code != 0
        assert 'Error: unknown FPGA: iCE40-FAKE' in result.output

        # apio build --fpga iCE40-FAKE --size 8k
        result = clirunner.invoke(cmd_build, [
            '--fpga', 'iCE40-FAKE', '--size', '8k'])
        assert result.exit_code != 0
        assert 'Error: unknown FPGA: iCE40-FAKE' in result.output

        # apio build --board icezum --fpga iCE40-FAKE
        result = clirunner.invoke(cmd_build, [
            '--board', 'icezum', '--fpga', 'iCE40-FAKE'])
        assert result.exit_code != 0
        assert 'Error: unknown FPGA: iCE40-FAKE' in result.output


def test_build_init(clirunner, configenv):
    with clirunner.isolated_filesystem():
        configenv()

        # apio init --board icezum
        result = clirunner.invoke(cmd_init, ['--board', 'icezum'])
        assert result.exit_code == 0
        assert 'Creating apio.ini file ...' in result.output
        assert 'has been successfully created!' in result.output

        # apio build
        result = clirunner.invoke(cmd_build)
        assert result.exit_code != 0

        # apio build --board icezum
        result = clirunner.invoke(cmd_build, ['--board', 'icestick'])
        assert result.exit_code != 0
        assert 'Info: ignore apio.ini board' in result.output

        # apio build --fpga iCE40-HX1K-VQ100
        result = clirunner.invoke(cmd_build, ['--fpga', 'iCE40-HX1K-VQ100'])
        assert result.exit_code != 0
        assert 'Info: ignore apio.ini board' in result.output

        # apio build --type lp --size 8k --pack cm225:4k
        result = clirunner.invoke(cmd_build, [
            '--type', 'lp', '--size', '8k', '--pack', 'cm225:4k'])
        assert result.exit_code != 0
        assert 'Info: ignore apio.ini board' in result.output

        # apio build --type lp --size 8k
        result = clirunner.invoke(cmd_build, ['--type', 'lp', '--size', '8k'])
        assert result.exit_code != 0
        assert 'Info: ignore apio.ini board' in result.output
        assert 'Error: insufficient arguments: missing pack' in result.output



# -*- coding: utf-8 -*-
# -- This file is part of the Apio project
# -- (C) 2016-2019 FPGAwars
# -- Author Jesús Arroyo
# -- Licence GPLv2

import click

from pkg_resources import get_distribution

from apio.util import get_pypi_latest_version


@click.command("upgrade")
@click.pass_context
def cli(ctx):
    """Check the latest Apio version."""

    current_version = get_distribution("apio").version
    latest_version = get_pypi_latest_version()

    if latest_version is None:
        ctx.exit(1)

    if latest_version == current_version:
        click.secho(
            "You're up-to-date!\nApio {} is currently the "
            "newest version available.".format(latest_version),
            fg="green",
        )
    else:
        click.secho(
            "You're not updated\nPlease execute "
            "`pip install -U apio` to upgrade.",
            fg="yellow",
        )



"""DOC: TODO"""
# -*- coding: utf-8 -*-
# -- This file is part of the Apio project
# -- (C) 2016-2019 FPGAwars
# -- Author Jesús Arroyo
# -- Licence GPLv2

import json
from pathlib import Path
from os.path import isfile, isdir
import click
import semantic_version


from apio import util


class Profile:
    """Class for managing the apio profile file"""

    def __init__(self):

        # ---- Set the default parameters
        # -- Apio default config mode
        self.config = {"exe": "default", "verbose": 0}

        self.labels = {"exe": "Executable", "verbose": "Verbose"}

        # Apio settings
        self.settings = {}

        # -- Installed package versions
        self.packages = {}

        # -- Get the profile path
        self._profile_path = str(Path(util.get_home_dir()) / "profile.json")

        print(f"(DEBUG) Profile path: {self._profile_path}")
        print(f"(DEBUG) Home_dir: {util.get_home_dir()}")

        # -- Read the profile from file
        self.load()

    def installed_version(self, name, version):
        """DOC: TODO"""

        if name in self.packages:
            pkg_version = self.get_package_version(name)
            pkg_version = self._convert_old_version(pkg_version)
            version = self._convert_old_version(version)
            return semantic_version.Version(
                pkg_version
            ) == semantic_version.Version(version)
        return None

    @staticmethod
    def _convert_old_version(version):
        # Convert old versions to new format
        try:
            ver = int(version)
            version = "1.{}.0".format(ver)
        except ValueError:
            pass
        return version

    def check_exe_default(self):
        """DOC: todo"""

        return self.config.get("exe", "") == "default"

    def add_package(self, name, version):
        """DOC: todo"""

        self.packages[name] = {"version": version}

    def add_setting(self, key, value):
        """DOC: todo"""

        self.settings[key] = value

    def add_config(self, key, value):
        """DOC: todo"""

        if self.config.get(key, None) != value:
            self.config[key] = value
            self.save()
            click.secho(
                "{0} mode updated: {1}".format(
                    self.labels.get(key, ""), value
                ),
                fg="green",
            )
        else:
            click.secho(
                "{0} mode already {1}".format(self.labels.get(key, ""), value),
                fg="yellow",
            )

    def remove_package(self, name):
        """DOC: todo"""

        if name in self.packages.keys():
            del self.packages[name]

    def get_verbose_mode(self):
        """DOC: todo"""

        return int(self.config.get("verbose", False))

    def get_package_version(self, name, release_name=""):
        """DOC: todo"""

        version = "0.0.0"
        if name in self.packages:
            version = self.packages.get(name).get("version")
        elif release_name:
            dir_name = util.get_package_dir(release_name)
            if isdir(dir_name):
                filepath = util.safe_join(dir_name, "package.json")
                try:
                    with open(filepath, "r") as json_file:
                        tmp_data = json.load(json_file)
                        if "version" in tmp_data.keys():
                            version = tmp_data.get("version")
                except Exception:
                    pass
        return version

    def load(self):
        """Load the profile from the file"""

        # -- Check if the file exist
        if isfile(self._profile_path):

            # -- Open the profile file
            with open(self._profile_path, "r") as profile:

                # -- Read the profile file
                self._load_profile(profile)

    def _load_profile(self, profile):
        """Read the profile file
        profile: file descriptor
        """

        # -- Process the json file
        data = json.load(profile)

        # -- Add the configuration object
        if "config" in data.keys():
            self.config = data.get("config")

            if "exe" not in self.config.keys():
                self.config["exe"] = "default"

            if "verbose" not in self.config.keys():
                self.config["verbose"] = 0

        # -- Add the settings object
        if "settings" in data.keys():
            self.settings = data.get("settings")

        # -- Add the packages version object
        if "packages" in data.keys():
            self.packages = data.get("packages")

        else:
            self.packages = data  # Backward compatibility

    def save(self):
        """DOC: todo"""

        util.mkdir(self._profile_path)
        with open(self._profile_path, "w") as profile:
            data = {
                "config": self.config,
                "settings": self.settings,
                "packages": self.packages,
            }
            json.dump(data, profile, indent=4, sort_keys=True)

    def list(self):
        """DOC: todo"""

        for key in self.config:
            click.secho(
                "{0} mode: {1}".format(
                    self.labels.get(key, ""), self.config.get(key, "")
                ),
                fg="yellow",
            )



# -*- coding: utf-8 -*-
# -- This file is part of the Apio project
# -- (C) 2016-2019 FPGAwars
# -- Author Jesús Arroyo
# -- Licence GPLv2

import click

from apio.managers.installer import Installer
from apio.resources import Resources
from apio.profile import Profile

platforms = [
    "linux_x86_64",
    "linux_i686",
    "linux_armv7l",
    "linux_aarch64",
    "windows",
    "darwin",
]


@click.command("uninstall")
@click.pass_context
@click.argument("packages", nargs=-1)
@click.option("-a", "--all", is_flag=True, help="Uninstall all packages.")
@click.option(
    "-l", "--list", is_flag=True, help="List all installed packages."
)
@click.option(
    "-p",
    "--platform",
    type=click.Choice(platforms),
    metavar="platform",
    help="Set the platform [{}] (Advanced).".format(", ".join(platforms)),
)
def cli(ctx, packages, all, list, platform):
    """Uninstall packages."""

    if packages:
        _uninstall(packages, platform)
    elif all:  # pragma: no cover
        packages = Profile().packages
        _uninstall(packages, platform)
    elif list:
        Resources(platform).list_packages(installed=True, notinstalled=False)
    else:
        click.secho(ctx.get_help())


def _uninstall(packages, platform):
    if click.confirm("Do you want to continue?"):
        for package in packages:
            Installer(package, platform, checkversion=False).uninstall()
    else:
        click.secho("Abort!", fg="red")



"""Open source ecosystem for open FPGA boards"""
# -*- coding: utf-8 -*-
# -- This file is part of the Apio project
# -- (C) 2016-2019 FPGAwars
# -- Author Jesús Arroyo
# -- Licence GPLv2

# --------------------------------------------
# - Information for the Distribution package
# --------------------------------------------

VERSION = (0, 8, '0.post1')
__version__ = ".".join([str(s) for s in VERSION])

__title__ = "apio"
__description__ = "Open source ecosystem for open FPGA boards"
__url__ = "https://github.com/FPGAwars/apio"

__author__ = "Jesús Arroyo Torrens"
__email__ = "jesus.arroyo.torrens@gmail.com"

__license__ = "GPLv2"

# Enable this flag to load data from /etc/apio.json file
# Used in apio-debian distribution
LOAD_CONFIG_DATA = False



"""DOC: TODO"""
# -*- coding: utf-8 -*-
# -- This file is part of the Apio project
# -- (C) 2016-2019 FPGAwars
# -- Author Jesús Arroyo
# -- Licence GPLv2

import sys
import click
import requests

requests.packages.urllib3.disable_warnings()


def api_request(command, organization="FPGAwars"):
    """Perform a request of data to Github, through its API"""

    result = None
    req = None

    # -- Create the URL for accesing the github API
    cmd_url = f"https://api.github.com/repos/{organization}/{command}"

    # -- Do the request!
    try:
        req = requests.get(cmd_url)
        result = req.json()
        req.raise_for_status()

    # -- There is a connection problem
    except requests.exceptions.ConnectionError as exc:
        error_message = str(exc)
        if "NewConnectionError" in error_message:
            click.secho(
                "Error: could not connect to GitHub API.\n"
                "Check your internet connection and try again",
                fg="red",
            )
        else:
            click.secho(error_message, fg="red")
        sys.exit(1)

    # -- There is another error
    except Exception as exc:
        click.secho("Error: " + str(exc), fg="red")
        sys.exit(1)

    # -- In any case, close the request
    finally:
        if req:
            req.close()

    # -- Return the response object from github
    if result is None:
        click.secho("Error: wrong data from GitHub API", fg="red")
        sys.exit(1)

    return result



from apio.commands.sim import cli as cmd_sim


def test_sim(clirunner, configenv):
    with clirunner.isolated_filesystem():
        configenv()
        result = clirunner.invoke(cmd_sim, ['--board', 'icezum'])
        assert result.exit_code != 0
        if result.exit_code == 1:
            assert 'apio install iverilog' in result.output



"""DOC: TODO"""
# -*- coding: utf-8 -*-
# -- This file is part of the Apio project
# -- (C) 2016-2018 FPGAwars
# -- Author Jesús Arroyo
# -- Licence GPLv2
# -- Derived from:
# ---- Platformio project
# ---- (C) 2014-2016 Ivan Kravets <me@ikravets.com>
# ---- Licence Apache v2

import sys
import os
import re
import json
import locale
import platform
import subprocess
from threading import Thread
from os.path import isdir, isfile, join, dirname, exists
from pathlib import Path

import click
import semantic_version
from serial.tools.list_ports import comports
import requests

from apio import LOAD_CONFIG_DATA

# ----------------------------------------
# -- Constants
# ----------------------------------------

# -- Packages names
# -- If you need to create new packages, you should
# -- define first the constants here
# --
OSS_CAD_SUITE = "oss-cad-suite"
SYSTEM = "system"
SCONS = "scons"
GTKWAVE = "gtkwave"
IVERILOG = "iverilog"
VERILATOR = "verilator"
YOSYS = "yosys"
ICE40 = "ice40"
ECP5 = "ecp5"
FUJPROG = "fujprog"
ICESPROG = "icesprog"
DFU = "dfu"

# -- Name of the subfolder to store de executable files
BIN = "bin"

# -- Folder names. They are built from the
# -- packages names
OSS_CAD_SUITE_FOLDER = f"tools-{OSS_CAD_SUITE}"
SYSTEM_FOLDER = f"tools-{SYSTEM}"
SCONS_FOLDER = f"tool-{SCONS}"
GTKWAVE_FOLDER = f"tool-{GTKWAVE}"
IVERILOG_FOLDER = f"toolchain-{IVERILOG}"
VERILATOR_FOLDER = f"toolchain-{VERILATOR}"
YOSYS_FOLDER = f"toolchain-{YOSYS}"
ICE40_FOLDER = f"toolchain-{ICE40}"
ECP5_FOLDER = f"toolchain-{ECP5}"
FUJPROG_FOLDER = f"toolchain-{FUJPROG}"
ICESPROG_FOLDER = f"toolchain-{ICESPROG}"
DFU_FOLDER = f"toolchain-{DFU}"

requests.packages.urllib3.disable_warnings()


# Python3 compat
if sys.version_info > (3, 0):
    unicode = str


class ApioException(Exception):
    """DOC: TODO"""

    MESSAGE = None

    def __str__(self):  # pragma: no cover
        if self.MESSAGE:
            return self.MESSAGE.format(*self.args)

        return Exception.__str__(self)


class AsyncPipe(Thread):  # pragma: no cover
    """DOC: TODO"""

    def __init__(self, outcallback=None):
        Thread.__init__(self)
        self.outcallback = outcallback

        self._fd_read, self._fd_write = os.pipe()
        self._pipe_reader = os.fdopen(self._fd_read)
        self._buffer = []

        self.start()

    def get_buffer(self):
        """DOC: TODO"""

        return self._buffer

    def fileno(self):
        """DOC: TODO"""

        return self._fd_write

    def run(self):
        """DOC: TODO"""

        for line in iter(self._pipe_reader.readline, ""):
            line = line.strip()
            self._buffer.append(line)
            if self.outcallback:
                self.outcallback(line)
        self._pipe_reader.close()

    def close(self):
        """DOC: TODO"""

        os.close(self._fd_write)
        self.join()


def get_systype():
    """DOC: TODO"""

    type_ = platform.system().lower()
    arch = platform.machine().lower()
    if type_ == "windows":
        arch = "amd64" if platform.architecture()[0] == "64bit" else "x86"
    return "%s_%s" % (type_, arch) if arch else type_


try:
    codepage = locale.getdefaultlocale()[1]
    if "darwin" in get_systype():
        UTF = True
    else:
        UTF = codepage.lower().find("utf") >= 0
except Exception:
    # Incorrect locale implementation, assume the worst
    UTF = False


def unicoder(string):
    """Make sure a Unicode string is returned"""
    if isinstance(string, unicode):
        return string

    if isinstance(string, str):
        return decoder(string)

    return unicode(decoder(string))


def decoder(string):
    """DOC: TODO"""

    if UTF:
        try:
            return string.decode("utf-8")
        except Exception:
            return string.decode(codepage)
    return string.decode(codepage)


def safe_join(*paths):
    """Join paths in a Unicode-safe way"""
    try:
        return join(*paths)
    except UnicodeDecodeError:
        npaths = ()
        for path in paths:
            npaths += (unicoder(path),)
        return join(*npaths)


def _get_config_data():
    """Return the configuration data located in the /etc/apio.json file
    Only for Debian Distribution. It will return None otherwise"""

    # Default value
    _config_data = None

    # If the LOAD_CONFIG_DATA flag is set
    # Read the /etc/apio.json file and return its contents
    # as an object
    if LOAD_CONFIG_DATA:
        filepath = safe_join(os.sep, "etc", "apio.json")
        if isfile(filepath):
            with open(filepath, "r") as file:
                # Load the JSON file
                _config_data = json.loads(file.read())

    return _config_data


# Global variable with the configuration data
# Only for Debian distributions
config_data = _get_config_data()


def _get_projconf_option_dir(name, default=None):
    """Return the project option with the given name
    These options are place either on environment variables or
    into the /etc/apio.json file in the case of debian distributions

    All the APIO environment variables have the prefix "APIO_"

    Project options:

    * home_dir : APIO home directory
    """

    # -- Get the full name of the environment variable
    _env_name = f"APIO_{name.upper()}"

    # -- Check if the environment variable
    # -- is defined
    if _env_name in os.environ:

        # -- Read the value of the environmental variable
        _env_value = os.getenv(_env_name)

        # -- On window systems the environmental variables can
        # -- include the quotes (""). But not in Linux
        # -- If there are quotes, remove them
        if _env_value.startswith('"') and _env_value.endswith('"'):
            _env_value = _env_value[1:-1]

        # -- Debug: Print the environment variable (without quotes)
        # print(f"DEBUG: {_env_name}: {_env_value}")

        return _env_value

    # -- Check if the config option is defined in the
    # -- configuration file (Only Debian systems)
    if config_data and _env_name in config_data.keys():

        # -- Return the value of the option
        return config_data.get(_env_name)

    # -- Return the default home_dir
    return default


def get_home_dir():
    """Get the APIO Home dir. This is the apio folder where the profle is
    located and the packages installed. The APIO Home dir can be set in the
    APIO_HOME_DIR environment varible or in the /etc/apio.json file (in
    Debian). If not set, the user_HOME/.apio folder is used by default:
    Ej. Linux:  /home/obijuan/.apio
    If the folders does not exist, they are created
    It returns a list with all the folders
    """

    # -- Get the APIO_HOME_DIR env variable
    # -- It returns None if it was not defined
    apio_home_dir_env = _get_projconf_option_dir("home_dir")

    # -- Get the home dir. It is what the APIO_HOME_DIR env variable
    # -- says, or the default folder if None
    if apio_home_dir_env:
        home_dir = Path(apio_home_dir_env)
    else:
        home_dir = Path.home() / ".apio"

    # -- Create the folders if they do not exist
    try:
        home_dir.mkdir(parents=True, exist_ok=True)
    except PermissionError:
        click.secho(f"Error: no usable home directory {home_dir}", fg="red")
        sys.exit(1)

    # Return the home_dir as a string
    # In the future it should return the path object
    return str(home_dir)


def get_package_dir(pkg_name):
    """Return the APIO package dir of a given package
    Packages are installed in the following folder:
    * Default: $APIO_HOME_DIR/packages
    * $APIO_PKG_DIR/packages: if the APIO_PKG_DIR env variable is set
    * Return a String
    """

    # -- Get the apio home dir:
    apio_home_dir = get_home_dir()

    # -- Get the APIO_PKG_DIR env variable
    # -- It returns None if it was not defined
    apio_pkg_dir_env = _get_projconf_option_dir("pkg_dir")

    # -- Get the pkg base dir. It is what the APIO_PKG_DIR env variable
    # -- says, or the default folder if None
    if apio_pkg_dir_env:
        pkg_home_dir = Path(apio_pkg_dir_env)

    # -- Default value
    else:
        pkg_home_dir = Path(apio_home_dir)

    # -- Create the package folder
    package_dir = pkg_home_dir / "packages" / pkg_name

    # -- Return the folder if it exists
    if package_dir.exists():
        return str(package_dir)

    # -- Show an error message (for debugging)
    # click.secho(f"Folder does not exists: {package_dir}", fg="red")
    # sys.exit(1)

    # -- Return a null string if the folder does not exist
    return ""


def get_project_dir():
    """DOC: TODO"""

    return os.getcwd()


def call(cmd):
    """Execute the given command from the installed apio packages"""

    # -- Set the PATH environment variable for finding the
    # -- executables on the apio package folders first
    setup_environment()

    # -- Execute the command from the shell
    result = subprocess.call(cmd, shell=True)

    # -- Command not found
    if result == 127:
        message = f"ERROR. Comand not found!: {cmd}"
        click.secho(message, fg="red")

    return result


def setup_environment():
    """Set the environment variables and the system PATH"""

    # --- Get the table with the paths of all the apio packages
    base_dir = get_base_dir()

    # --- Get the table with the paths of all the executables
    # --- of the apio packages
    bin_dir = get_bin_dir_table(base_dir)

    # --- Set the system env. variables
    set_env_variables(base_dir, bin_dir)

    return bin_dir


def set_env_variables(base_dir, bin_dir):
    """Set the environment variables"""

    # -- Get the current system PATH
    path = os.environ["PATH"]

    # -- Add the packages to the path. The first packages added
    # -- have the lowest priority. The latest the highest

    # -- Add the gtkwave to the path if installed,
    # -- but only for windows platforms
    if platform.system() == "Windows":

        # -- Gtkwave package is installed
        if bin_dir[GTKWAVE] != "":
            path = os.pathsep.join([bin_dir.get(GTKWAVE), path])

    # -- Add the binary folders of the installed packages
    # -- to the path, except for the OSS_CAD_SUITE package
    for pack in base_dir:
        if base_dir[pack] != "" and pack != OSS_CAD_SUITE:
            path = os.pathsep.join([bin_dir[pack], path])

    # -- Add the OSS_CAD_SUITE package to the path
    # -- if installed (Maximum priority)
    if base_dir[OSS_CAD_SUITE] != "":

        # -- Get the lib folder (where the shared libraries are located)
        oss_cad_suite_lib = str(Path(base_dir[OSS_CAD_SUITE]) / "lib")

        # -- Add the lib folder
        path = os.pathsep.join([oss_cad_suite_lib, path])
        path = os.pathsep.join([bin_dir[OSS_CAD_SUITE], path])

    # Add the virtual python environment to the path
    os.environ["PATH"] = path

    # print(f" get_bin_dir(): {get_bin_dir()}")

    # -- DEBUG
    print()
    print(f"PATH: {os.environ['PATH']}")
    print()

    # Add other environment variables

    os.environ["IVL"] = str(Path(base_dir[OSS_CAD_SUITE]) / "lib" / "ivl")

    os.environ["ICEBOX"] = str(
        Path(base_dir[OSS_CAD_SUITE]) / "share" / "icebox"
    )

    os.environ["TRELLIS"] = str(
        Path(base_dir[OSS_CAD_SUITE]) / "share" / "trellis"
    )

    os.environ["YOSYS_LIB"] = str(
        Path(base_dir[OSS_CAD_SUITE]) / "share" / "yosys"
    )


def resolve_packages(packages, installed_packages, spec_packages):
    """Check the given packages.
    * Check that they are installed
    * Check that the versions are ok"""

    # --- Get the table with the paths of all the apio packages
    base_dir = get_base_dir()

    # --- Get the table with the paths of all the executables
    # --- of the apio packages
    bin_dir = get_bin_dir_table(base_dir)

    # -- Check packages
    check = True
    for package in packages:

        version = installed_packages.get(package, {}).get("version", "")

        spec_version = spec_packages.get(package, "")

        check &= check_package(
            package, version, spec_version, bin_dir.get(package)
        )

    # -- Load packages
    if check:

        # --- Set the system env. variables
        set_env_variables(base_dir, bin_dir)

    return check


def get_base_dir():
    """Return the table with the local paths of the apio packages
    installed on the system. If the packages is not installed,
    the path is ''
    """

    # -- Create the table:
    # --  Package Name  :  Folder (string)
    base_dir = {
        OSS_CAD_SUITE: get_package_dir(OSS_CAD_SUITE_FOLDER),
        SCONS: get_package_dir(SCONS_FOLDER),
        YOSYS: get_package_dir(YOSYS_FOLDER),
        ICE40: get_package_dir(ICE40_FOLDER),
        ECP5: get_package_dir(ECP5_FOLDER),
        IVERILOG: get_package_dir(IVERILOG_FOLDER),
        VERILATOR: get_package_dir(VERILATOR_FOLDER),
        GTKWAVE: get_package_dir(GTKWAVE_FOLDER),
        FUJPROG: get_package_dir(FUJPROG_FOLDER),
        ICESPROG: get_package_dir(ICESPROG_FOLDER),
        DFU: get_package_dir(DFU_FOLDER),
        # -- Obsolete packages
        SYSTEM: get_package_dir(SYSTEM_FOLDER),
    }

    return base_dir


def get_bin_dir_table(base_dir):
    """Return a table with the package name and the folder were
    the executable files are stored
    * Input: Table with the package base_dir
    """

    bin_dir = {
        OSS_CAD_SUITE: str(Path(base_dir.get(OSS_CAD_SUITE)) / BIN),
        SCONS: str(Path(base_dir.get(SCONS)) / "script"),
        YOSYS: str(Path(base_dir.get(YOSYS)) / BIN),
        ICE40: str(Path(base_dir.get(ICE40)) / BIN),
        ECP5: str(Path(base_dir.get(ECP5)) / BIN),
        IVERILOG: str(Path(base_dir.get(IVERILOG)) / BIN),
        VERILATOR: str(Path(base_dir.get(VERILATOR)) / BIN),
        GTKWAVE: str(Path(base_dir.get(GTKWAVE)) / BIN),
        FUJPROG: str(Path(base_dir.get(FUJPROG)) / BIN),
        ICESPROG: str(Path(base_dir.get(ICESPROG)) / BIN),
        DFU: str(Path(base_dir.get(DFU)) / BIN),
        # -- Obsolete package
        SYSTEM: str(Path(base_dir.get(SYSTEM)) / BIN),
    }

    return bin_dir


def check_package(name, version, spec_version, path):
    """Check if the given package is installed
    * name: Package name
    * path: path where the binary files of the package are stored
    """

    # Apio package 'gtkwave' only exists for Windows.
    # Linux and MacOS user must install the native GTKWave.
    if name == "gtkwave" and platform.system() != "Windows":
        return True

    # Check package path
    if not isdir(path):
        show_package_path_error(name)
        show_package_install_instructions(name)
        return False

    # Check package version
    if not check_package_version(version, spec_version):
        show_package_version_warning(name, version, spec_version)
        show_package_install_instructions(name)
        return False

    return True


def check_package_version(version, spec_version):
    """DOC: TODO"""

    spec = semantic_version.SimpleSpec(spec_version)
    try:
        semver = semantic_version.Version(version)
    except ValueError:
        return False

    return semver in spec


def show_package_version_warning(name, version, spec_version):
    """DOC: TODO"""

    message = (
        "Warning: package '{0}' version {1}\n"
        "does not match the semantic version {2}"
    ).format(name, version, spec_version)
    click.secho(message, fg="yellow")


def show_package_path_error(name):
    """DOC: TODO"""

    message = f"Error: package '{name}' is not installed"
    click.secho(message, fg="red")


def show_package_install_instructions(name):
    """DOC: TODO"""

    if config_data and _check_apt_get():  # /etc/apio.json file exists
        click.secho(
            f"Please run:\n   apt-get install apio-{name}",
            fg="yellow",
        )
    else:
        click.secho(f"Please run:\n   apio install {name}", fg="yellow")


def _check_apt_get():
    """Check if apio can be installed through apt-get"""
    check = False
    if "TESTING" not in os.environ:
        result = exec_command(["dpkg", "-l", "apio"])
        if result and result.get("returncode") == 0:
            match = re.findall(r"rc\s+apio", result.get("out")) + re.findall(
                r"ii\s+apio", result.get("out")
            )
            check = len(match) > 0
    return check


def get_package_version(name, profile):
    """DOC: TODO"""

    version = ""
    if name in profile.packages:
        version = profile.packages.get(name).get("version")
    return version


def get_package_spec_version(name, resources):
    """DOC: TODO"""

    spec_version = ""
    if name in resources.distribution.get("packages"):
        spec_version = resources.distribution.get("packages").get(name)
    return spec_version


def change_filemtime(path, time):
    """DOC: TODO"""

    os.utime(path, (time, time))


def exec_command(*args, **kwargs):  # pragma: no cover
    """DOC: TODO"""

    result = {"out": None, "err": None, "returncode": None}

    default = dict(
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        shell=False,
    )
    default.update(kwargs)
    kwargs = default

    try:
        proc = subprocess.Popen(*args, **kwargs)
        result["out"], result["err"] = proc.communicate()
        result["returncode"] = proc.returncode
    except KeyboardInterrupt:
        click.secho("Aborted by user", fg="red")
        sys.exit(1)
    except Exception as exc:
        click.secho(str(exc), fg="red")
        sys.exit(1)
    finally:
        for std in ("stdout", "stderr"):
            if isinstance(kwargs[std], AsyncPipe):
                kwargs[std].close()

    _parse_result(kwargs, result)

    return result


def _parse_result(kwargs, result):
    for std in ("stdout", "stderr"):
        if isinstance(kwargs[std], AsyncPipe):
            result[std[3:]] = "\n".join(kwargs[std].get_buffer())

    for k, value in result.items():
        if value and isinstance(value, unicode):
            result[k].strip()


def get_pypi_latest_version():
    """DOC: TODO"""

    req = None
    version = None
    try:
        req = requests.get("https://pypi.python.org/pypi/apio/json")
        version = req.json().get("info").get("version")
        req.raise_for_status()
    except requests.exceptions.ConnectionError as exc:
        error_message = str(exc)
        if "NewConnectionError" in error_message:
            click.secho(
                "Error: could not connect to Pypi.\n"
                "Check your internet connection and try again",
                fg="red",
            )
        else:
            click.secho(error_message, fg="red")
    except Exception as exc:
        click.secho("Error: " + str(exc), fg="red")
    finally:
        if req:
            req.close()
    return version


def get_folder(folder):
    """DOC: TODO"""

    return safe_join(dirname(__file__), folder)


def mkdir(path):
    """DOC: TODO"""

    path = dirname(path)
    if not exists(path):
        try:
            os.makedirs(path)
        except OSError:
            pass


def check_dir(_dir):
    """Check if the given path is a folder. If no path is given
    the current path is used"""

    # -- If no path is given, get the current working directory
    if _dir is None:
        _dir = os.getcwd()

    # -- Check if the path is a file or a folder
    if isfile(_dir):

        # -- It is a file! Error! Exit!
        click.secho(
            f"Error: project directory is already a file: {_dir}", fg="red"
        )

        sys.exit(1)

    # -- If the folder does not exist....
    if not exists(_dir):

        # -- Warning
        click.secho(f"Warning: The path does not exist: {_dir}", fg="yellow")

        # -- Create the folder
        click.secho(f"Creating folder: {_dir}")
        os.makedirs(_dir)

    # -- Return the path
    return _dir


def command(function):
    """Command decorator"""

    def decorate(*args, **kwargs):
        exit_code = 1
        try:
            exit_code = function(*args, **kwargs)
        except Exception as exc:
            if str(exc):
                click.secho("Error: " + str(exc), fg="red")
        finally:
            return exit_code

    return decorate


def get_serial_ports():
    """DOC: TODO"""

    result = []

    for port, description, hwid in comports():
        if not port:
            continue
        if "VID:PID" in hwid:
            result.append(
                {"port": port, "description": description, "hwid": hwid}
            )

    return result


def get_tinyprog_meta():
    """DOC: TODO"""

    _command = join(get_bin_dir(), "tinyprog")
    result = exec_command([_command, "--pyserial", "--meta"])
    try:
        out = unicoder(result.get("out", ""))
        if out:
            return json.loads(out)
    except Exception as exc:
        print(exc)
        return []
    return []


def get_bin_dir():
    """DOC: TODO"""

    candidate = dirname(sys.modules["__main__"].__file__)
    # Windows + virtualenv = 💩
    # In this case the main file is: venv/Scripts/apio.exe/__main__.py!
    # This is not good because venv/Scripts/apio.exe is not a directory
    # So here we go with the workaround:
    if candidate.endswith(".exe"):
        return dirname(candidate)

    return candidate


def get_python_version():
    """Return a string with the python version"""

    return f"{sys.version_info[0]}.{sys.version_info[1]}"



"""DOC: TODO"""
# -*- coding: utf-8 -*-
# -- This file is part of the Apio project
# -- (C) 2016-2021 FPGAwars
# -- Author Jesús Arroyo
# -- Licence GPLv2

import sys
import re
import shutil
from os import makedirs, remove, rename
from os.path import isfile, isdir, basename
import click


from apio import util
from apio.api import api_request
from apio.resources import Resources
from apio.profile import Profile

from apio.managers.downloader import FileDownloader
from apio.managers.unpacker import FileUnpacker


class Installer:
    """Installer. Class with methods for installing and managing
    packages"""

    def __init__(self, package, platform="", force=False, checkversion=True):
        """Class initialization. Parameters:
        * package:  Package name to manage/install. It can have a prefix with
                    the version. Ex. "system@1.1.2"
        """

        # Parse version. The following attributes are used:
        #  * Installer.package: Package name (without version)
        #  * Installer.version: Package version (or None)
        if "@" in package:
            split = package.split("@")
            self.package = split[0]
            self.version = split[1]
        else:
            self.package = package
            self.version = None

        # -- Attribute Installer.force_install
        # -- Force installation or not
        self.force_install = force

        # -- Installer.package_dir: path were the packages are stored
        # -- Ex. /home/obijuan/.apio/packages
        self.packages_dir = ""

        # -- Get all the resources for the given platform
        # -- Some resources depend on the platform (like the packages)
        # -- but some others don't (like the boards)
        self.resources = Resources(platform)

        self.profile = Profile()

        # -- Folder name were the packages are stored
        dirname = "packages"

        # -- If the package is known...
        # --(It is defined in the resources/packages.json file)
        if self.package in self.resources.packages:

            # -- Store the package dir
            self.packages_dir = util.safe_join(util.get_home_dir(), dirname)

            # Get the data of the given package
            data = self.resources.packages.get(self.package)

            # Get the information about the valid versions
            distribution = self.resources.distribution

            # Get the spectec package version
            self.spec_version = distribution.get("packages").get(self.package)

            # Get the package name (from resources/package.json file)
            self.package_name = data.get("release").get("package_name")

            # Get the extension given to the toolchain. Tipically tar.gz
            self.extension = data.get("release").get("extension")

            # Get the current platform
            platform = platform or self._get_platform()

            # Check if the version is ok (It is only done if the
            # checkversion flag has been activated)
            if checkversion:

                # Check version. The filename is read from the
                # repostiroy
                valid_version = self._get_valid_version(
                    data.get("repository").get("name"),
                    data.get("repository").get("organization"),
                    data.get("release").get("tag_name"),
                )

                # No valid version found
                if not valid_version:
                    # Error
                    click.secho("Error: no valid version found", fg="red")
                    sys.exit(1)

                # Store the valid version
                self.version = valid_version

                # Get the plaform_os name
                # e.g., [linux_x86_64, linux]
                platform_os = platform.split("_")[0]

                # Build the URLs for downloading the package
                self.download_urls = [
                    {
                        "url": self.get_download_url(data, platform),
                        "platform": platform,
                    },
                    {
                        "url": self.get_download_url(data, platform_os),
                        "platform": platform_os,
                    },
                ]
        # -- The package is kwnown but the version is not correct
        else:

            if self.package in self.profile.packages and checkversion is False:
                self.packages_dir = util.safe_join(
                    util.get_home_dir(), dirname
                )
                self.package_name = "toolchain-" + package

        # -- If the Installer.package_dir is empty, is because the package
        # -- was not known. Abort!
        if self.packages_dir == "":
            click.secho(
                "Error: no such package '{}'".format(self.package), fg="red"
            )
            sys.exit(1)

    def get_download_url(self, data, platform):
        """DOC: TODO"""

        compressed_name = data.get("release").get("compressed_name")
        self.compressed_name = compressed_name.replace(
            "%V", self.version
        ).replace("%P", platform)
        uncompressed_name = data.get("release").get("uncompressed_name")
        self.uncompressed_name = uncompressed_name.replace(
            "%V", self.version
        ).replace("%P", platform)

        tarball = self._get_tarball_name(self.compressed_name, self.extension)

        download_url = self._get_download_url(
            data.get("repository").get("name"),
            data.get("repository").get("organization"),
            data.get("release").get("tag_name").replace("%V", self.version),
            tarball,
        )

        return download_url

    def install(self):
        """Install the current package in the set in the Installer Object"""

        # -- Warning if the package has been marked as obsolete
        if self.package in self.resources.obsolete_pkgs:

            # -- Get the string with the new package to use instead
            # -- of the obsolete one (if available)
            if self.resources.obsolete_pkgs[self.package] != "":
                new_package = self.resources.obsolete_pkgs[self.package]
                new_package_msg = f"Use {new_package} instead\n"
            else:
                new_package_msg = ""

            click.secho(
                f"Warning: Package {self.package} is obsolete. "
                f"Will be removed in the future. "
                f"{new_package_msg}",
                fg="yellow",
            )

        click.secho(f"Installing {self.package} package:", fg="cyan")

        # -- Create the apio package folder, if it does not exit
        if not isdir(self.packages_dir):
            makedirs(self.packages_dir)
        assert isdir(self.packages_dir)

        # -- The first step is downloading the package
        # -- This variable stores the path to tha packages
        dlpath = None

        try:
            # Try full platform
            platform_download_url = self.download_urls[0].get("url")
            dlpath = self._download(platform_download_url)

        # -- There is no write access to the package folder
        except IOError as exc:
            click.secho(
                "Warning: permission denied in packages directory", fg="yellow"
            )
            click.secho(str(exc), fg="red")

        # --
        except Exception:
            # Try os name
            dlpath = self._install_os_package(platform_download_url)

        # -- Second step: Install downloaded package
        self._install_package(dlpath)

        # Rename unpacked dir to package dir
        self._rename_unpacked_dir()

    def _install_os_package(self, platform_download_url):
        os_download_url = self.download_urls[1].get("url")
        if platform_download_url != os_download_url:
            click.secho(
                "Warning: full platform does not match: {}\
                ".format(
                    self.download_urls[0].get("platform")
                ),
                fg="yellow",
            )
            click.secho(
                "         Trying OS name: {}\
                ".format(
                    self.download_urls[1].get("platform")
                ),
                fg="yellow",
            )
            try:
                return self._download(os_download_url)
            except Exception as exc:
                click.secho("Error: {}".format(str(exc)), fg="red")
        else:
            click.secho(
                "Error: package not availabe for this platform", fg="red"
            )
        return None

    def _install_package(self, dlpath):
        if dlpath:
            package_dir = util.safe_join(self.packages_dir, self.package_name)
            if isdir(package_dir):
                shutil.rmtree(package_dir)
            if self.uncompressed_name:
                self._unpack(dlpath, self.packages_dir)
            else:
                self._unpack(
                    dlpath,
                    util.safe_join(self.packages_dir, self.package_name),
                )

            remove(dlpath)
            self.profile.add_package(self.package, self.version)
            self.profile.save()
            click.secho(
                """Package \'{}\' has been """
                """successfully installed!""".format(self.package),
                fg="green",
            )

    def _rename_unpacked_dir(self):
        if self.uncompressed_name:
            unpack_dir = util.safe_join(
                self.packages_dir, self.uncompressed_name
            )
            package_dir = util.safe_join(self.packages_dir, self.package_name)
            if isdir(unpack_dir):
                rename(unpack_dir, package_dir)

    def uninstall(self):
        """DOC: TODO"""

        if isdir(util.safe_join(self.packages_dir, self.package_name)):
            click.echo(
                "Uninstalling %s package:"
                % click.style(self.package, fg="cyan")
            )
            shutil.rmtree(util.safe_join(self.packages_dir, self.package_name))
            click.secho(
                """Package \'{}\' has been """
                """successfully uninstalled!""".format(self.package),
                fg="green",
            )
        else:
            util.show_package_path_error(self.package)
        self.profile.remove_package(self.package)
        self.profile.save()

    @staticmethod
    def _get_platform():
        return util.get_systype()

    @staticmethod
    def _get_download_url(name, organization, tag, tarball):
        url = "https://github.com/{0}/{1}/releases/download/{2}/{3}".format(
            organization, name, tag, tarball
        )
        return url

    @staticmethod
    def _get_tarball_name(name, extension):
        tarball = "{0}.{1}".format(name, extension)
        return tarball

    def _get_valid_version(self, rel_name, organization, tag_name):
        """Get the latest valid version. It can be the latest vesion available
        in github or a specific version"""

        # Get the data from the Github api
        # Ex of the object obtained in this URL:
        # https://api.github.com/repos/FPGAwars/tools-oss-cad-suite/releases
        releases = api_request(f"{rel_name}/releases", organization)

        if self.version:

            # Find required version via @
            if not util.check_package_version(self.version, self.spec_version):

                util.show_package_version_warning(
                    self.package, self.version, self.spec_version
                )
                sys.exit(1)

            return self._find_required_version(
                releases, tag_name, self.version
            )

        # Find latest version release
        return self._find_latest_version(releases, tag_name, self.spec_version)

    def _find_required_version(self, releases, tag_name, req_v):
        for release in releases:
            if "tag_name" in release:
                tag = tag_name.replace("%V", req_v)
                if tag == release.get("tag_name"):
                    prerelease = release.get("prerelease", False)
                    if prerelease and not self.force_install:
                        click.secho(
                            "Warning: "
                            + req_v
                            + " is"
                            + " a pre-release. Use --force to install",
                            fg="yellow",
                        )
                        sys.exit(1)
                    return req_v
        return None

    @staticmethod
    def _find_latest_version(releases, tag_name, spec_v):
        for release in releases:
            if "tag_name" in release:
                pattern = tag_name.replace("%V", "(?P<v>.*?)") + "$"
                match = re.search(pattern, release.get("tag_name"))
                if match:
                    prerelease = release.get("prerelease", False)
                    if not prerelease:
                        version = match.group("v")
                        if util.check_package_version(version, spec_v):
                            return version
        return None

    def _download(self, url):
        # Note: here we check only for the version of locally installed
        # packages. For this reason we don't say what's the installation
        # path.
        if (
            not self.profile.installed_version(self.package, self.version)
            or self.force_install
        ):
            filed = FileDownloader(url, self.packages_dir)
            filepath = filed.get_filepath()
            click.secho("Download " + basename(filepath))
            try:
                filed.start()
            except KeyboardInterrupt:
                if isfile(filepath):
                    remove(filepath)
                click.secho("Abort download!", fg="red")
                sys.exit(1)
            return filepath

        click.secho(
            "Already installed. Version {0}".format(
                self.profile.get_package_version(self.package)
            ),
            fg="yellow",
        )
        return None

    @staticmethod
    def _unpack(pkgpath, pkgdir):
        fileu = FileUnpacker(pkgpath, pkgdir)
        return fileu.start()



"""DOC: TODO"""
# -*- coding: utf-8 -*-
# -- This file is part of the Apio project
# -- (C) 2016-2019 FPGAwars
# -- Author Jesús Arroyo
# -- Licence GPLv2

import os
import re
import sys
import time
import datetime
from os.path import isfile
from pathlib import Path

import click
import pkg_resources
import semantic_version

from apio import util
from apio.managers.arguments import process_arguments
from apio.managers.arguments import format_vars
from apio.managers.system import System
from apio.profile import Profile
from apio.resources import Resources


class SCons:
    """Class for managing the scons tools"""

    def __init__(self, project_dir=""):
        """Initialization:
        * project_dir: path where the sources are located
          If not given, the curent working dir is used
        """

        # -- Read the apio profile file
        self.profile = Profile()

        # -- Read the apio resources
        self.resources = Resources()

        # -- Project path is given
        if project_dir is not None:

            # Check if it is a correct folder
            # (or create a new one)
            project_dir = util.check_dir(project_dir)

            # Change to that folder
            os.chdir(project_dir)

    @util.command
    def clean(self, args):
        """DOC: TODO"""

        try:
            __, __, arch = process_arguments(args, self.resources)

        # -- No architecture given: Uses ice40 as default
        except Exception:
            arch = "ice40"

        # --Clean the project: run scons -c (with aditional arguments)
        return self.run("-c", arch=arch)

    @util.command
    def verify(self, args):
        """DOC: TODO"""

        __, __, arch = process_arguments(args, self.resources)
        return self.run(
            "verify",
            arch=arch,
            packages=["oss-cad-suite"],
        )

    @util.command
    def lint(self, args):
        """DOC: TODO"""

        __, __, arch = process_arguments(None, self.resources)
        var = format_vars(
            {
                "all": args.get("all"),
                "top": args.get("top"),
                "nowarn": args.get("nowarn"),
                "warn": args.get("warn"),
                "nostyle": args.get("nostyle"),
            }
        )
        return self.run(
            "lint",
            var,
            arch=arch,
            packages=["oss-cad-suite"],
        )

    @util.command
    def sim(self):
        """DOC: TODO"""

        __, __, arch = process_arguments(None, self.resources)
        return self.run(
            "sim",
            arch=arch,
            packages=["iverilog", "oss-cad-suite", "gtkwave"],
        )

    @util.command
    def build(self, args):
        """Build the circuit"""

        # -- Split the arguments
        var, board, arch = process_arguments(args, self.resources)

        # -- Execute scons!!!
        # -- The packages to check are passed
        return self.run(
            "build",
            var,
            board,
            arch,
            packages=["oss-cad-suite"],
        )

    @util.command
    def time(self, args):
        """DOC: TODO"""

        var, board, arch = process_arguments(args, self.resources)
        return self.run(
            "time", var, board, arch, packages=["oss-cad-suite", arch]
        )

    @util.command
    def upload(self, args, serial_port, ftdi_id, sram, flash):
        """Upload the circuit to the board"""

        # -- Split the arguments
        var, board, arch = process_arguments(args, self.resources)

        # -- Get the programmer for that board
        programmer = self.get_programmer(
            board, serial_port, ftdi_id, sram, flash
        )

        var += [f"prog={programmer}"]

        return self.run(
            "upload",
            var,
            board,
            arch,
            packages=["oss-cad-suite"],
        )

    def get_programmer(self, board, ext_serial, ext_ftdi_id, sram, flash):
        """DOC: TODO"""

        programmer = ""

        if board:

            board_data = self.resources.boards.get(board)

            # Check platform
            self.check_platform(board_data)

            # Check pip packages
            self.check_pip_packages(board_data)

            # Serialize programmer command
            programmer = self.serialize_programmer(board_data, sram, flash)

            # Replace USB vendor id
            if "${VID}" in programmer:
                vid = board_data.get("usb").get("vid")
                programmer = programmer.replace("${VID}", vid)

            # Replace USB product id
            if "${PID}" in programmer:
                pid = board_data.get("usb").get("pid")
                programmer = programmer.replace("${PID}", pid)

            # Replace FTDI index
            if "${FTDI_ID}" in programmer:
                self.check_usb(board, board_data)
                ftdi_id = self.get_ftdi_id(board, board_data, ext_ftdi_id)
                programmer = programmer.replace("${FTDI_ID}", ftdi_id)

            # TinyFPGA BX board is not detected in MacOS HighSierra
            if "tinyprog" in board_data and "darwin" in util.get_systype():
                # In this case the serial check is ignored
                return "tinyprog --libusb --program"

            # Replace Serial port
            if "${SERIAL_PORT}" in programmer:
                self.check_usb(board, board_data)
                device = self.get_serial_port(board, board_data, ext_serial)
                programmer = programmer.replace("${SERIAL_PORT}", device)

        return programmer

    @staticmethod
    def check_platform(board_data):
        """DOC: TODO"""

        if "platform" not in board_data:
            return

        platform = board_data.get("platform")
        current_platform = util.get_systype()
        if platform != current_platform:
            # Incorrect platform
            if platform == "linux_armv7l":
                raise Exception("incorrect platform: RPI2 or RPI3 required")
            raise Exception("incorrect platform {0}".format(platform))

    def check_pip_packages(self, board_data):
        """Check if the corresponding pip package with the programmer
        has already been installed. In the case of an apio package
        it is just ignored
        """

        # -- Get the programmer object for the given board
        prog_info = board_data.get("programmer")

        # -- Get the programmer information (from the type)
        # -- Command, arguments, pip package, etc...
        prog_data = self.resources.programmers.get(prog_info.get("type"))

        # -- Get all the pip packages from the distribution
        all_pip_packages = self.resources.distribution.get("pip_packages")

        # -- Get the name of the pip package of the current programmer,
        # -- if any (The programmer maybe in a pip package or an apio package)
        pip_packages = prog_data.get("pip_packages") or []

        # -- Check if pip package was installed
        # -- In case of an apio package it is just ignored
        for pip_pkg in pip_packages:
            try:
                # Check pip_package version
                spec = semantic_version.Spec(all_pip_packages.get(pip_pkg, ""))
                pkg_version = pkg_resources.get_distribution(pip_pkg).version
                version = semantic_version.Version(pkg_version)
                if not spec.match(version):
                    click.secho(
                        "Error: '{}' ".format(pip_pkg)
                        + "version ({}) ".format(version)
                        + "does not match {}".format(spec),
                        fg="red",
                    )
                    click.secho(
                        "Please run:\n"
                        "   pip install -U apio[{}]".format(pip_pkg),
                        fg="yellow",
                    )
                    raise Exception
            except pkg_resources.DistributionNotFound:
                click.secho(
                    "Error: '{}' is not installed".format(pip_pkg), fg="red"
                )
                click.secho(
                    "Please run:\n"
                    "   pip install -U apio[{}]".format(pip_pkg),
                    fg="yellow",
                )
                raise Exception
            try:
                # Check pip_package itself
                __import__(pip_pkg)
            except Exception as exc:
                # Exit if a package is not working
                python_version = util.get_python_version()
                message = "'{}' not compatible with ".format(pip_pkg)
                message += "Python {}".format(python_version)
                message += "\n       {}".format(exc)
                raise Exception(message)

    def serialize_programmer(self, board_data, sram, flash):
        """DOC: TODO"""

        prog_info = board_data.get("programmer")
        content = self.resources.programmers.get(prog_info.get("type"))

        programmer = content.get("command")

        # Add args
        if content.get("args"):
            programmer += " {}".format(content.get("args"))

        # Add extra args
        if prog_info.get("extra_args"):
            programmer += " {}".format(prog_info.get("extra_args"))

        # Enable SRAM programming
        if sram:
            # Only for iceprog programmer
            if programmer.startswith("iceprog"):
                programmer += " -S"

        if flash:
            # Only for ujprog programmer
            if programmer.startswith("ujprog"):
                programmer = programmer.replace("SRAM", "FLASH")

        return programmer

    @staticmethod
    def check_usb(board, board_data):
        """DOC: TODO"""

        if "usb" not in board_data:
            raise Exception("Missing board configuration: usb")

        usb_data = board_data.get("usb")
        hwid = "{0}:{1}".format(usb_data.get("vid"), usb_data.get("pid"))
        found = False
        for usb_device in System().get_usb_devices():
            if usb_device.get("hwid") == hwid:
                found = True
                break

        if not found:
            # Board not connected
            if "tinyprog" in board_data:
                click.secho(
                    "Activate bootloader by pressing the reset button",
                    fg="yellow",
                )
            raise Exception("board " + board + " not connected")

    def get_serial_port(self, board, board_data, ext_serial_port):
        """DOC: TODO"""

        # Search Serial port by USB id
        device = self._check_serial(board, board_data, ext_serial_port)
        if device is None:
            # Board not connected
            raise Exception("board " + board + " not connected")
        return device

    def _check_serial(self, board, board_data, ext_serial_port):
        """DOC: TODO"""

        if "usb" not in board_data:
            raise Exception("Missing board configuration: usb")

        usb_data = board_data.get("usb")
        hwid = "{0}:{1}".format(usb_data.get("vid"), usb_data.get("pid"))

        # Match the discovered serial ports
        serial_ports = util.get_serial_ports()
        if len(serial_ports) == 0:
            # Board not available
            raise Exception("board " + board + " not available")
        for serial_port_data in serial_ports:
            port = serial_port_data.get("port")
            if ext_serial_port and ext_serial_port != port:
                # If the --device options is set but it doesn't match
                # the detected port, skip the port.
                continue
            if hwid.lower() in serial_port_data.get("hwid").lower():
                if "tinyprog" in board_data and not self._check_tinyprog(
                    board_data, port
                ):
                    # If the board uses tinyprog use its port detection
                    # to double check the detected port.
                    # If the port is not detected, skip the port.
                    continue
                # If the hwid and the description pattern matches
                # with the detected port return the port.
                return port
        return None

    @staticmethod
    def _check_tinyprog(board_data, port):
        """DOC: TODO"""

        desc_pattern = "^" + board_data.get("tinyprog").get("desc") + "$"
        for tinyprog_meta in util.get_tinyprog_meta():
            tinyprog_port = tinyprog_meta.get("port")
            tinyprog_name = tinyprog_meta.get("boardmeta").get("name")
            if port == tinyprog_port and re.match(desc_pattern, tinyprog_name):
                # If the port is detected and it matches the pattern
                return True
        return False

    def get_ftdi_id(self, board, board_data, ext_ftdi_id):
        """DOC: TODO"""

        # Search device by FTDI id
        ftdi_id = self._check_ftdi(board, board_data, ext_ftdi_id)
        if ftdi_id is None:
            # Board not connected
            raise Exception("board " + board + " not connected")
        return ftdi_id

    @staticmethod
    def _check_ftdi(board, board_data, ext_ftdi_id):
        """DOC: TODO"""

        if "ftdi" not in board_data:
            raise Exception("Missing board configuration: ftdi")

        desc_pattern = "^" + board_data.get("ftdi").get("desc") + "$"

        # Match the discovered FTDI chips
        ftdi_devices = System().get_ftdi_devices()
        if len(ftdi_devices) == 0:
            # Board not available
            raise Exception("board " + board + " not available")
        for ftdi_device in ftdi_devices:
            index = ftdi_device.get("index")
            if ext_ftdi_id and ext_ftdi_id != index:
                # If the --device options is set but it doesn't match
                # with the detected index, skip the port.
                continue
            if re.match(desc_pattern, ftdi_device.get("description")):
                # If matches the description pattern
                # return the index for the FTDI device.
                return index
        return None

    def run(self, command, variables=[], board=None, arch=None, packages=[]):
        """Executes scons for building"""

        # -- Check if in the current project a custom SConstruct file
        # is being used. We fist build the full name (with the full path)
        scon_file = Path(util.get_project_dir()) / "SConstruct"

        # -- If the SConstruct file does NOT exist, we use the one provided by
        # -- apio, which is located in the resources/arch/ folder
        if not isfile(scon_file):

            # -- This is the default SConstruct file
            resources = Path(util.get_folder("resources"))
            default_scons_file = resources / arch / "SConstruct"

            # -- It is passed to scons using the flag -f default_scons_file
            variables += ["-f", f"{default_scons_file}"]

        else:
            # -- We are using our custom SConstruct file
            click.secho("Info: use custom SConstruct file")

        # -- Check the configuration mode
        if self.profile.check_exe_default():

            # Run on `default` config mode

            # -- Check if the necessary packages are installed
            if not util.resolve_packages(
                packages,
                self.profile.packages,
                self.resources.distribution.get("packages"),
            ):
                # Exit if a package is not installed
                raise Exception
        else:
            click.secho("Info: native config mode")

        # -- Execute scons
        return self._execute_scons(command, variables, board)

    def _execute_scons(self, command, variables, board):
        """Execute the scons builder"""

        terminal_width, _ = click.get_terminal_size()
        start_time = time.time()

        if command in ("build", "upload", "time"):
            if board:
                processing_board = board
            else:
                processing_board = "custom board"
            click.echo(
                "[%s] Processing %s"
                % (
                    datetime.datetime.now().strftime("%c"),
                    click.style(processing_board, fg="cyan", bold=True),
                )
            )
            click.secho("-" * terminal_width, bold=True)

        scons_command = ["scons"] + ["-Q", command] + variables
        if self.profile.get_verbose_mode() > 0:
            click.secho(
                "Executing: {}".format(
                    " ".join(util.exec_command(scons_command))
                )
            )

        # -- Execute the scons builder
        result = util.exec_command(
            scons_command,
            stdout=util.AsyncPipe(self._on_stdout),
            stderr=util.AsyncPipe(self._on_stderr),
        )

        # -- Print result
        exit_code = result.get("returncode")
        is_error = exit_code != 0
        summary_text = " Took %.2f seconds " % (time.time() - start_time)
        half_line = "=" * int(((terminal_width - len(summary_text) - 10) / 2))
        click.echo(
            "%s [%s]%s%s"
            % (
                half_line,
                (
                    click.style(" ERROR ", fg="red", bold=True)
                    if is_error
                    else click.style("SUCCESS", fg="green", bold=True)
                ),
                summary_text,
                half_line,
            ),
            err=is_error,
        )

        return exit_code

    @staticmethod
    def _on_stdout(line):
        fgcol = "green" if "is up to date" in line else None
        click.secho(line, fg=fgcol)

    @staticmethod
    def _on_stderr(line):
        if "%|" in line and "100%|" not in line:
            # Remove previous line for tqdm progress bar
            cursor_up = "\033[F"
            erase_line = "\033[K"
            sys.stdout.write(cursor_up + erase_line)
        fgcol = "red" if "error" in line.lower() else "yellow"
        click.secho(line, fg=fgcol)



# -*- coding: utf-8 -*-
# -- This file is part of the Apio project
# -- (C) 2016-2019 FPGAwars
# -- Author Jesús Arroyo
# -- Licence GPLv2

import sys
import click

from apio.managers.scons import SCons

# Python3 compat
if sys.version_info > (3, 0):
    unicode = str


@click.command("time")
@click.pass_context
@click.option(
    "-b", "--board", type=unicode, metavar="board", help="Set the board."
)
@click.option("--fpga", type=unicode, metavar="fpga", help="Set the FPGA.")
@click.option(
    "--size", type=unicode, metavar="size", help="Set the FPGA type (1k/8k)."
)
@click.option(
    "--type", type=unicode, metavar="type", help="Set the FPGA type (hx/lp)."
)
@click.option(
    "--pack", type=unicode, metavar="package", help="Set the FPGA package."
)
@click.option(
    "-p",
    "--project-dir",
    type=unicode,
    metavar="path",
    help="Set the target directory for the project.",
)
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    help="Show the entire output of the command.",
)
@click.option(
    "--verbose-yosys",
    is_flag=True,
    help="Show the yosys output of the command.",
)
@click.option(
    "--verbose-pnr", is_flag=True, help="Show the pnr output of the command."
)
def cli(
    ctx,
    board,
    fpga,
    pack,
    type,
    size,
    project_dir,
    verbose,
    verbose_yosys,
    verbose_pnr,
):
    """Bitstream timing analysis."""

    # Run scons
    exit_code = SCons(project_dir).time(
        {
            "board": board,
            "fpga": fpga,
            "size": size,
            "type": type,
            "pack": pack,
            "verbose": {
                "all": verbose,
                "yosys": verbose_yosys,
                "pnr": verbose_pnr,
            },
        }
    )
    ctx.exit(exit_code)



# -*- coding: utf-8 -*-
# -- This file is part of the Apio project
# -- (C) 2016-2019 FPGAwars
# -- Author Jesús Arroyo
# -- Licence GPLv2

import os
import subprocess
from os.path import isfile
from pathlib import Path
import click

from apio import util
from apio.profile import Profile
from apio.resources import Resources

platform = util.get_systype()

FTDI_INSTALL_DRIVER_INSTRUCTIONS = """
   FTDI driver installation:
   Usage instructions

      1. Connect the FTDI FPGA board
      2. Select (Interface 0)
      3. Replace driver by "libusbK"
      4. Reconnect the board
      5. Check `apio system --lsftdi`
"""

FTDI_UNINSTALL_DRIVER_INSTRUCTIONS = """
   FTDI driver uninstallation:
   Usage instructions

      1. Find the FPGA USB Device
      2. Right click
      3. Select "Uninstall"
      4. Accept the dialog
"""

SERIAL_INSTALL_DRIVER_INSTRUCTIONS = """
   Serial driver installation:
   Usage instructions

      1. Connect the Serial FPGA board
      2. Install the driver
      3. Reconnect the board
      4. Check `apio system --lsserial`
"""

SERIAL_UNINSTALL_DRIVER_INSTRUCTIONS = """
   Serial driver uninstallation:
   Usage instructions

      1. Find the FPGA USB Device
      2. Right click
      3. Select "Uninstall"
      4. Accept the dialog
"""


class Drivers:  # pragma: no cover

    # -- The driver installation on linux consist of copying the rule files
    # -- to the /etc/udev/rules.d folder

    # -- FTDI source rules file paths
    resources = Path(util.get_folder("resources"))
    ftdi_rules_local_path = resources / "80-fpga-ftdi.rules"

    # -- Target rule file
    ftdi_rules_system_path = "/etc/udev/rules.d/80-fpga-ftdi.rules"

    # -- It was the target in older versions of apio
    old_ftdi_rules_system_path = "/etc/udev/rules.d/80-icestick.rules"

    # Serial rules files paths
    serial_rules_local_path = resources / "80-fpga-serial.rules"
    serial_rules_system_path = "/etc/udev/rules.d/80-fpga-serial.rules"

    # Driver to restore: mac os
    driver_c = ""

    def __init__(self) -> None:
        self.profile = None
        self.name = None
        self.version = None
        self.spec_version = None

    def ftdi_enable(self):
        """Enable the FTDI driver. It depends on the platform"""

        # -- Driver enabling on Linux
        if "linux" in platform:
            return self._ftdi_enable_linux()

        # -- Driver enabling on MAC
        if "darwin" in platform:
            self._setup_darwin()
            return self._ftdi_enable_darwin()

        # -- Driver enabling on Windows
        if "windows" in platform:
            self._setup_windows()
            return self._ftdi_enable_windows()
        return None

    def ftdi_disable(self):
        """Disable the FTDI driver. It depends on the platform"""

        # -- Linux platforms
        if "linux" in platform:
            return self._ftdi_disable_linux()

        # -- MAC
        if "darwin" in platform:
            self._setup_darwin()
            return self._ftdi_disable_darwin()

        # -- Windows
        if "windows" in platform:
            self._setup_windows()
            return self._ftdi_disable_windows()

        return None

    def serial_enable(self):
        """Enable the Serial driver. It depends on the platform"""

        if "linux" in platform:
            return self._serial_enable_linux()

        if "darwin" in platform:
            self._setup_darwin()
            return self._serial_enable_darwin()

        if "windows" in platform:
            self._setup_windows()
            return self._serial_enable_windows()
        return None

    def serial_disable(self):
        """Disable the Serial driver. It depends on the platform"""

        if "linux" in platform:
            return self._serial_disable_linux()

        if "darwin" in platform:
            self._setup_darwin()
            return self._serial_disable_darwin()

        if "windows" in platform:
            self._setup_windows()
            return self._serial_disable_windows()
        return None

    def pre_upload(self):
        """Operations to do before uploading a design
        Only for mac platforms"""

        if "darwin" in platform:
            self._setup_darwin()
            self._pre_upload_darwin()

    def post_upload(self):
        """Operations to do after uploading a design
        Only for mac platforms"""

        if "darwin" in platform:
            self._setup_darwin()
            self._post_upload_darwin()

    def _setup_darwin(self):
        """Setup operation on Mac"""

        # -- Just read the profile file
        self.profile = Profile()

    def _setup_windows(self):
        """Setup operations on Windows"""

        # -- Read the Profile and Resources files
        profile = Profile()
        resources = Resources()

        # -- On windows the zadig driver installer should be
        # -- execute. Get the package version
        self.name = "drivers"
        self.version = util.get_package_version(self.name, profile)
        self.spec_version = util.get_package_spec_version(self.name, resources)

    def _ftdi_enable_linux(self):
        """Drivers enable on Linux. It copies the .rules file into
        the corresponding folder"""

        click.secho("Configure FTDI drivers for FPGA")

        # -- Check if the target rules file already exists
        if not isfile(self.ftdi_rules_system_path):

            # -- The file does not exist. Copy!
            # -- Execute the cmd: sudo cp src_file target_file
            subprocess.call(
                [
                    "sudo",
                    "cp",
                    str(self.ftdi_rules_local_path),
                    self.ftdi_rules_system_path,
                ]
            )

            # -- Execute the commands for reloading the udev system
            self._reload_rules()

            click.secho("FTDI drivers enabled", fg="green")
            click.secho("Unplug and reconnect your board", fg="yellow")
        else:
            click.secho("Already enabled", fg="yellow")

    def _ftdi_disable_linux(self):
        """Disable the FTDI drivers on linux"""

        # -- For disabling the FTDI driver the .rules files should be
        # -- removed from the /etc/udev/rules.d/ folder

        # -- Remove the old .rules files, if it exists
        if isfile(self.old_ftdi_rules_system_path):
            subprocess.call(["sudo", "rm", self.old_ftdi_rules_system_path])

        # -- Remove the .rules file, if it exists
        if isfile(self.ftdi_rules_system_path):
            click.secho("Revert FTDI drivers configuration")

            # -- Execute the sudo rm rules_file command
            subprocess.call(["sudo", "rm", self.ftdi_rules_system_path])

            # -- # -- Execute the commands for reloading the udev system
            self._reload_rules()

            click.secho("FTDI drivers disabled", fg="green")
            click.secho("Unplug and reconnect your board", fg="yellow")
        else:
            click.secho("Already disabled", fg="yellow")

    def _serial_enable_linux(self):
        """Serial drivers enable on Linux"""

        click.secho("Configure Serial drivers for FPGA")

        # -- Check if the target rules file already exists
        if not isfile(self.serial_rules_system_path):

            # -- Add the user to the dialout group for
            # -- having access to the serial port
            group_added = self._add_dialout_group()

            # -- The file does not exist. Copy!
            # -- Execute the cmd: sudo cp src_file target_file
            subprocess.call(
                [
                    "sudo",
                    "cp",
                    str(self.serial_rules_local_path),
                    self.serial_rules_system_path,
                ]
            )

            # -- Execute the commands for reloading the udev system
            self._reload_rules()

            click.secho("Serial drivers enabled", fg="green")
            click.secho("Unplug and reconnect your board", fg="yellow")
            if group_added:
                click.secho(
                    "Restart your machine to enable the dialout group",
                    fg="yellow",
                )
        else:
            click.secho("Already enabled", fg="yellow")

    def _serial_disable_linux(self):
        """Disable the serial driver on Linux"""

        # -- For disabling the serial driver the corresponding .rules file
        # -- should be removed, it it exists
        if isfile(self.serial_rules_system_path):
            click.secho("Revert Serial drivers configuration")

            # -- Execute the sudo rm rule_file cmd
            subprocess.call(["sudo", "rm", self.serial_rules_system_path])

            # -- Execute the commands for reloading the udev system
            self._reload_rules()
            click.secho("Serial drivers disabled", fg="green")
            click.secho("Unplug and reconnect your board", fg="yellow")
        else:
            click.secho("Already disabled", fg="yellow")

    @staticmethod
    def _reload_rules():
        """Execute the commands for reloading the udev system"""

        # -- These are Linux commands that should be executed on
        # -- the shell
        subprocess.call(["sudo", "udevadm", "control", "--reload-rules"])
        subprocess.call(["sudo", "udevadm", "trigger"])
        subprocess.call(["sudo", "service", "udev", "restart"])

    @staticmethod
    def _add_dialout_group():
        """Add the current user to the dialout group on Linux systems"""

        # -- This operation is needed for granting access to the serial port

        # -- Get the current groups of the user
        groups = subprocess.check_output("groups")

        # -- If it does not belong to the dialout group, add it!!
        if "dialout" not in groups.decode():

            # -- Command for adding the user to the dialout group
            subprocess.call("sudo usermod -a -G dialout $USER", shell=True)
            return True
        return None

    def _ftdi_enable_darwin(self):
        # Check homebrew
        brew = subprocess.call("which brew > /dev/null", shell=True)
        if brew != 0:
            click.secho("Error: homebrew is required", fg="red")
        else:
            click.secho("Enable FTDI drivers for FPGA")
            subprocess.call(["brew", "update"])
            self._brew_install("libffi")
            self._brew_install("libftdi")
            self.profile.add_setting("macos_ftdi_drivers", True)
            self.profile.save()
            click.secho("FTDI drivers enabled", fg="green")

    def _ftdi_disable_darwin(self):
        click.secho("Disable FTDI drivers configuration")
        self.profile.add_setting("macos_ftdi_drivers", False)
        self.profile.save()
        click.secho("FTDI drivers disabled", fg="green")

    def _serial_enable_darwin(self):
        # Check homebrew
        brew = subprocess.call("which brew > /dev/null", shell=True)
        if brew != 0:
            click.secho("Error: homebrew is required", fg="red")
        else:
            click.secho("Enable Serial drivers for FPGA")
            subprocess.call(["brew", "update"])
            self._brew_install("libffi")
            self._brew_install("libusb")
            # self._brew_install_serial_drivers()
            click.secho("Serial drivers enabled", fg="green")

    @staticmethod
    def _serial_disable_darwin():
        click.secho("Disable Serial drivers configuration")
        click.secho("Serial drivers disabled", fg="green")

    @staticmethod
    def _brew_install(package):
        subprocess.call(["brew", "install", "--force", package])
        subprocess.call(["brew", "unlink", package])
        subprocess.call(["brew", "link", "--force", package])

    @staticmethod
    def _brew_install_serial_drivers():
        subprocess.call(
            [
                "brew",
                "tap",
                "mengbo/ch340g-ch34g-ch34x-mac-os-x-driver",
                "https://github.com/mengbo/ch340g-ch34g-ch34x-mac-os-x-driver",
            ]
        )
        subprocess.call(
            ["brew", "cask", "install", "wch-ch34x-usb-serial-driver"]
        )

    def _pre_upload_darwin(self):
        if self.profile.settings.get("macos_ftdi_drivers", False):
            # Check and unload the drivers
            driver_a = "com.FTDI.driver.FTDIUSBSerialDriver"
            driver_b = "com.apple.driver.AppleUSBFTDI"
            if self._check_ftdi_driver_darwin(driver_a):
                subprocess.call(["sudo", "kextunload", "-b", driver_a])
                self.driver_c = driver_a
            elif self._check_ftdi_driver_darwin(driver_b):
                subprocess.call(["sudo", "kextunload", "-b", driver_b])
                self.driver_c = driver_b

    def _post_upload_darwin(self):
        if self.profile.settings.get("macos_ftdi_drivers", False):
            # Restore previous driver configuration
            if self.driver_c:
                subprocess.call(["sudo", "kextload", "-b", self.driver_c])

    @staticmethod
    def _check_ftdi_driver_darwin(driver):
        return driver in str(subprocess.check_output(["kextstat"]))

    def _ftdi_enable_windows(self):
        drivers_base_dir = util.get_package_dir("tools-drivers")
        drivers_bin_dir = util.safe_join(drivers_base_dir, "bin")
        drivers_share_dir = util.safe_join(drivers_base_dir, "share")
        zadig_ini_path = util.safe_join(drivers_share_dir, "zadig.ini")
        zadig_ini = "zadig.ini"

        try:
            if util.check_package(
                self.name, self.version, self.spec_version, drivers_bin_dir
            ):
                click.secho("Launch drivers configuration tool")
                click.secho(FTDI_INSTALL_DRIVER_INSTRUCTIONS, fg="yellow")
                # Copy zadig.ini
                with open(zadig_ini, "w") as ini_file:
                    with open(zadig_ini_path, "r") as local_ini_file:
                        ini_file.write(local_ini_file.read())

                result = util.exec_command(
                    util.safe_join(drivers_bin_dir, "zadig.exe")
                )
                click.secho("FTDI drivers configuration finished", fg="green")
            else:
                result = 1
        except Exception as exc:
            click.secho("Error: " + str(exc), fg="red")
            result = 1
        finally:
            # Remove zadig.ini
            if isfile(zadig_ini):
                os.remove(zadig_ini)

        if not isinstance(result, int):
            result = result.get("returncode")
        return result

    @staticmethod
    def _ftdi_disable_windows():
        click.secho("Launch device manager")
        click.secho(FTDI_UNINSTALL_DRIVER_INSTRUCTIONS, fg="yellow")

        result = util.exec_command("mmc devmgmt.msc")
        return result.get("returncode")

    def _serial_enable_windows(self):
        drivers_base_dir = util.get_package_dir("tools-drivers")
        drivers_bin_dir = util.safe_join(drivers_base_dir, "bin")

        try:
            if util.check_package(
                self.name, self.version, self.spec_version, drivers_bin_dir
            ):
                click.secho("Launch drivers configuration tool")
                click.secho(SERIAL_INSTALL_DRIVER_INSTRUCTIONS, fg="yellow")
                result = util.exec_command(
                    util.safe_join(drivers_bin_dir, "serial_install.exe")
                )
                click.secho(
                    "Serial drivers configuration finished", fg="green"
                )
            else:
                result = 1
        except Exception as exc:
            click.secho("Error: " + str(exc), fg="red")
            result = 1

        if not isinstance(result, int):
            result = result.get("returncode")
        return result

    @staticmethod
    def _serial_disable_windows():
        click.secho("Launch device manager")
        click.secho(SERIAL_UNINSTALL_DRIVER_INSTRUCTIONS, fg="yellow")

        result = util.exec_command("mmc devmgmt.msc")
        return result.get("returncode")



# -*- coding: utf-8 -*-
# -- This file is part of the Apio project
# -- (C) 2016-2019 FPGAwars
# -- Author Jesús Arroyo
# -- Licence GPLv2

import sys
import click

from apio.managers.scons import SCons

# Python3 compat
if sys.version_info > (3, 0):
    unicode = str


@click.command("sim")
@click.pass_context
@click.option(
    "-p",
    "--project-dir",
    type=unicode,
    metavar="path",
    help="Set the target directory for the project.",
)
def cli(ctx, project_dir):
    """Launch the verilog simulation."""

    exit_code = SCons(project_dir).sim()
    ctx.exit(exit_code)



# -*- coding: utf-8 -*-
# -- This file is part of the Apio project
# -- (C) 2016-2019 FPGAwars
# -- Authors Jesús Arroyo, Miodrag Milanovic
# -- Licence GPLv2

import sys
import click

from apio.managers.scons import SCons

# Python3 compat
if sys.version_info > (3, 0):
    unicode = str


@click.command("lint")
@click.pass_context
@click.option(
    "-a",
    "--all",
    is_flag=True,
    help="Enable all warnings, including code style warnings.",
)
@click.option(
    "-t", "--top", type=unicode, metavar="top", help="Set top module."
)
@click.option("--nostyle", is_flag=True, help="Disable all style warnings.")
@click.option(
    "--nowarn",
    type=unicode,
    metavar="nowarn",
    help="Disable specific warning(s).",
)
@click.option(
    "--warn", type=unicode, metavar="warn", help="Enable specific warning(s)."
)
@click.option(
    "-p",
    "--project-dir",
    type=unicode,
    metavar="path",
    help="Set the target directory for the project.",
)
def cli(ctx, all, top, nostyle, nowarn, warn, project_dir):
    """Lint the verilog code."""

    exit_code = SCons(project_dir).lint(
        {
            "all": all,
            "top": top,
            "nostyle": nostyle,
            "nowarn": nowarn,
            "warn": warn,
        }
    )
    ctx.exit(exit_code)



from apio.commands.verify import cli as cmd_verify


def test_verify(clirunner, configenv):
    with clirunner.isolated_filesystem():
        configenv()
        result = clirunner.invoke(cmd_verify, ['--board', 'icezum'])
        assert result.exit_code != 0
        if result.exit_code == 1:
            assert 'apio install oss-cad-suite' in result.output



from apio.commands.boards import cli as cmd_boards


def test_boards(clirunner, validate_cliresult):
    result = clirunner.invoke(cmd_boards)
    validate_cliresult(result)


def test_boards_list(clirunner, validate_cliresult):
    result = clirunner.invoke(cmd_boards, ['--list'])
    validate_cliresult(result)


def test_boards_fpgas(clirunner, validate_cliresult):
    result = clirunner.invoke(cmd_boards, ['--fpga'])
    validate_cliresult(result)



# -*- coding: utf-8 -*-
# -- This file is part of the Apio project
# -- (C) 2016-2019 FPGAwars
# -- Author Jesús Arroyo
# -- Licence GPLv2

import click

from apio.resources import Resources


@click.command("boards")
@click.pass_context
@click.option(
    "-l", "--list", is_flag=True, help="List all supported FPGA boards."
)
@click.option(
    "-f", "--fpga", is_flag=True, help="List all supported FPGA chips."
)
# pylint: disable=redefined-builtin,
def cli(ctx, list, fpga):
    """Manage FPGA boards."""

    if list:
        Resources().list_boards()
    elif fpga:
        Resources().list_fpgas()
    else:
        click.secho(ctx.get_help())



# -*- coding: utf-8 -*-
# -- This file is part of the Apio project
# -- (C) 2016-2019 FPGAwars
# -- Author Jesús Arroyo
# -- Licence GPLv2

from os.path import isfile

import click

from apio.managers.project import Project


def process_arguments(args, resources):  # noqa
    # -- Check arguments
    if args is not None:
        var_board = args.get("board")
        var_arch = args.get("arch")
        var_fpga = args.get("fpga")
        var_size = args.get("size")
        var_type = args.get("type")
        var_pack = args.get("pack")
        var_idcode = args.get("idcode")
        var_verbose = args.get("verbose")
    else:
        var_board = None
        var_arch = None
        var_fpga = None
        var_size = None
        var_type = None
        var_pack = None
        var_idcode = None
        var_verbose = {}

    if var_board:
        if isfile("apio.ini"):
            click.secho("Info: ignore apio.ini board", fg="yellow")
        if var_board in resources.boards:
            fpga = resources.boards.get(var_board).get("fpga")
            if fpga in resources.fpgas:
                fpga_arch = resources.fpgas.get(fpga).get("arch")
                fpga_size = resources.fpgas.get(fpga).get("size")
                fpga_type = resources.fpgas.get(fpga).get("type")
                fpga_pack = resources.fpgas.get(fpga).get("pack")
                fpga_idcode = resources.fpgas.get(fpga).get("idcode")

                redundant_arguments = []
                contradictory_arguments = []

                if var_fpga:
                    if var_fpga in resources.fpgas:
                        if var_fpga == fpga:
                            # Redundant argument
                            redundant_arguments += ["fpga"]
                        else:
                            # Contradictory argument
                            contradictory_arguments += ["fpga"]
                    else:
                        # Unknown FPGA
                        raise Exception("unknown FPGA: {0}".format(var_fpga))

                if var_size:
                    if var_size == fpga_size:
                        # Redundant argument
                        redundant_arguments += ["size"]
                    else:
                        # Contradictory argument
                        contradictory_arguments += ["size"]

                if var_type:
                    if var_type == fpga_type:
                        # Redundant argument
                        redundant_arguments += ["type"]
                    else:
                        # Contradictory argument
                        contradictory_arguments += ["type"]

                if var_pack:
                    if var_pack == fpga_pack:
                        # Redundant argument
                        redundant_arguments += ["pack"]
                    else:
                        # Contradictory argument
                        contradictory_arguments += ["pack"]

                if var_idcode:
                    if var_idcode == fpga_idcode:
                        # Redundant argument
                        redundant_arguments += ["idcode"]
                    else:
                        # Contradictory argument
                        contradictory_arguments += ["idcode"]

                if redundant_arguments:
                    # Redundant argument
                    click.secho(
                        "Warning: redundant arguments: {}".format(
                            ", ".join(redundant_arguments)
                        ),
                        fg="yellow",
                    )

                if contradictory_arguments:
                    # Contradictory argument
                    raise Exception(
                        "contradictory arguments: {}".format(
                            ", ".join(contradictory_arguments)
                        )
                    )
            else:
                # Unknown FPGA
                pass
        else:
            # Unknown board
            raise Exception("unknown board: {0}".format(var_board))
    else:
        if var_fpga:
            if isfile("apio.ini"):
                click.secho("Info: ignore apio.ini board", fg="yellow")
            if var_fpga in resources.fpgas:
                fpga_arch = resources.fpgas.get(var_fpga).get("arch")
                fpga_size = resources.fpgas.get(var_fpga).get("size")
                fpga_type = resources.fpgas.get(var_fpga).get("type")
                fpga_pack = resources.fpgas.get(var_fpga).get("pack")
                fpga_idcode = resources.fpgas.get(var_fpga).get("idcode")

                redundant_arguments = []
                contradictory_arguments = []

                if var_size:
                    if var_size == fpga_size:
                        # Redundant argument
                        redundant_arguments += ["size"]
                    else:
                        # Contradictory argument
                        contradictory_arguments += ["size"]

                if var_type:
                    if var_type == fpga_type:
                        # Redundant argument
                        redundant_arguments += ["type"]
                    else:
                        # Contradictory argument
                        contradictory_arguments += ["type"]

                if var_pack:
                    if var_pack == fpga_pack:
                        # Redundant argument
                        redundant_arguments += ["pack"]
                    else:
                        # Contradictory argument
                        contradictory_arguments += ["pack"]

                if var_idcode:
                    if var_idcode == fpga_idcode:
                        # Redundant argument
                        redundant_arguments += ["idcode"]
                    else:
                        # Contradictory argument
                        contradictory_arguments += ["idcode"]

                if redundant_arguments:
                    # Redundant argument
                    click.secho(
                        "Warning: redundant arguments: {}".format(
                            ", ".join(redundant_arguments)
                        ),
                        fg="yellow",
                    )

                if contradictory_arguments:
                    # Contradictory argument
                    raise Exception(
                        "contradictory arguments: {}".format(
                            ", ".join(contradictory_arguments)
                        )
                    )
            else:
                # Unknown FPGA
                raise Exception("unknown FPGA: {0}".format(var_fpga))
        else:
            if var_size and var_type and var_pack and var_arch:
                if isfile("apio.ini"):
                    click.secho("Info: ignore apio.ini board", fg="yellow")
                fpga_arch = var_arch
                fpga_size = var_size
                fpga_type = var_type
                fpga_pack = var_pack
                fpga_idcode = var_idcode
            else:
                if not var_size and not var_type and not var_pack:
                    # No arguments: use apio.ini board
                    proj = Project()
                    proj.read()
                    if proj.board:
                        var_board = proj.board
                        if var_board in resources.boards:
                            fpga = resources.boards.get(var_board).get("fpga")
                            fpga_arch = resources.fpgas.get(fpga).get("arch")
                            fpga_size = resources.fpgas.get(fpga).get("size")
                            fpga_type = resources.fpgas.get(fpga).get("type")
                            fpga_pack = resources.fpgas.get(fpga).get("pack")
                            fpga_idcode = resources.fpgas.get(fpga).get(
                                "idcode"
                            )
                        else:
                            # Unknown board
                            raise Exception(
                                "unknown board: {0}".format(var_board)
                            )
                    else:
                        click.secho(
                            "Error: insufficient arguments: missing board",
                            fg="red",
                        )
                        click.secho(
                            "You have two options:\n"
                            + "  1) Execute your command with\n"
                            + "       `--board <boardname>`\n"
                            + "  2) Create an ini file using\n"
                            + "       `apio init --board <boardname>`",
                            fg="yellow",
                        )
                        raise Exception
                else:
                    if isfile("apio.ini"):
                        click.secho("Info: ignore apio.ini board", fg="yellow")
                    # Insufficient arguments
                    missing = []
                    if not var_size:
                        missing += ["size"]
                    if not var_type:
                        missing += ["type"]
                    if not var_pack:
                        missing += ["pack"]
                    raise Exception(
                        "insufficient arguments: missing {}".format(
                            ", ".join(missing)
                        )
                    )

    # -- Build Scons variables list
    variables = format_vars(
        {
            "fpga_arch": fpga_arch,
            "fpga_size": fpga_size,
            "fpga_type": fpga_type,
            "fpga_pack": fpga_pack,
            "fpga_idcode": fpga_idcode,
            "verbose_all": var_verbose.get("all"),
            "verbose_yosys": var_verbose.get("yosys"),
            "verbose_pnr": var_verbose.get("pnr"),
        }
    )

    return variables, var_board, fpga_arch


def format_vars(args):
    """Format the given vars in the form: 'flag=value'"""
    variables = []
    for key, value in args.items():
        if value:
            variables += ["{0}={1}".format(key, value)]
    return variables



from apio.commands.drivers import cli as cmd_drivers


def test_drivers(clirunner, validate_cliresult, configenv):
    with clirunner.isolated_filesystem():
        configenv()
        result = clirunner.invoke(cmd_drivers)
        validate_cliresult(result)



# -*- coding: utf-8 -*-
# -- This file is part of the Apio project
# -- (C) 2016-2019 FPGAwars
# -- Author Jesús Arroyo
# -- Licence GPLv2

import sys
import click

from apio.managers.scons import SCons

# Python3 compat
if sys.version_info > (3, 0):
    unicode = str


@click.command("clean")
@click.pass_context
@click.option(
    "-p",
    "--project-dir",
    type=unicode,
    metavar="path",
    help="Set the target directory for the project.",
)
@click.option(
    "-b", "--board", type=unicode, metavar="board", help="Set the board."
)
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    help="Show the entire output of the command.",
)
def cli(ctx, board, project_dir, verbose):
    """Clean the previous generated files."""
    exit_code = SCons(project_dir).clean(
        {"board": board, "verbose": {"all": verbose}}
    )
    ctx.exit(exit_code)



# -*- coding: utf-8 -*-

import pytest
from os import environ, getcwd, path
from click.testing import CliRunner


@pytest.fixture(scope='module')
def clirunner():
    return CliRunner()


@pytest.fixture(scope='session')
def validate_cliresult():
    def decorator(result):
        assert result.exit_code == 0
        assert not result.exception
        assert 'error' not in result.output.lower()
    return decorator


@pytest.fixture(scope='session')
def configenv():
    def decorator():
        cwd = path.join(getcwd(), ' ñ')
        environ['APIO_HOME_DIR'] = cwd
        environ['APIO_PKG_DIR'] = cwd
        environ['TESTING'] = ''
    return decorator


def pytest_addoption(parser):
    parser.addoption('--offline', action='store_true',
                     help='Run tests in offline mode')


@pytest.fixture
def offline(request):
    return request.config.getoption('--offline')



# -*- coding: utf-8 -*-
# -- This file is part of the Apio project
# -- (C) 2016-2019 FPGAwars
# -- Author Jesús Arroyo
# -- Licence GPLv2

import click

from apio.managers.drivers import Drivers


@click.command("drivers")
@click.pass_context
@click.option("--ftdi-enable", is_flag=True, help="Enable FTDI drivers.")
@click.option("--ftdi-disable", is_flag=True, help="Disable FTDI drivers.")
@click.option("--serial-enable", is_flag=True, help="Enable Serial drivers.")
@click.option("--serial-disable", is_flag=True, help="Disable Serial drivers.")
def cli(ctx, ftdi_enable, ftdi_disable, serial_enable, serial_disable):
    """Manage FPGA boards drivers."""

    exit_code = 0

    if ftdi_enable:  # pragma: no cover
        exit_code = Drivers().ftdi_enable()
    elif ftdi_disable:  # pragma: no cover
        exit_code = Drivers().ftdi_disable()
    elif serial_enable:  # pragma: no cover
        exit_code = Drivers().serial_enable()
    elif serial_disable:  # pragma: no cover
        exit_code = Drivers().serial_disable()
    else:
        click.secho(ctx.get_help())

    ctx.exit(exit_code)



from apio.commands.clean import cli as cmd_clean


def test_clean(clirunner, configenv):
    with clirunner.isolated_filesystem():
        configenv()
        result = clirunner.invoke(cmd_clean)
        assert result.exit_code != 0



from apio.commands.config import cli as cmd_config


def test_config(clirunner, validate_cliresult, configenv):
    with clirunner.isolated_filesystem():
        configenv()
        result = clirunner.invoke(cmd_config)
        validate_cliresult(result)


def test_config_list(clirunner, validate_cliresult, configenv):
    with clirunner.isolated_filesystem():
        configenv()
        result = clirunner.invoke(cmd_config, ['--list'])
        validate_cliresult(result)


def test_config_exe(clirunner, validate_cliresult, configenv):
    with clirunner.isolated_filesystem():
        configenv()
        result = clirunner.invoke(cmd_config, ['--exe', 'native'])
        validate_cliresult(result)



# -*- coding: utf-8 -*-
# -- This file is part of the Apio project
# -- (C) 2016-2019 FPGAwars
# -- Author Juan González, Jesús Arroyo
# -- Licence GPLv2

import sys
import json
from os.path import isfile
import click

try:
    import ConfigParser
except ImportError:
    import configparser as ConfigParser

from apio import util
from apio.resources import Resources

PROJECT_FILENAME = "apio.ini"


class Project:
    def __init__(self):
        self.board = None

    def create_sconstruct(self, project_dir="", arch=None, sayyes=False):
        """Creates a default SConstruct file"""

        project_dir = util.check_dir(project_dir)

        sconstruct_name = "SConstruct"
        sconstruct_path = util.safe_join(project_dir, sconstruct_name)
        local_sconstruct_path = util.safe_join(
            util.get_folder("resources"), arch, sconstruct_name
        )

        if isfile(sconstruct_path):
            # -- If sayyes, skip the question
            if sayyes:
                self._copy_sconstruct_file(
                    sconstruct_name, sconstruct_path, local_sconstruct_path
                )
            else:
                click.secho(
                    "Warning: {} file already exists".format(sconstruct_name),
                    fg="yellow",
                )

                if click.confirm("Do you want to replace it?"):
                    self._copy_sconstruct_file(
                        sconstruct_name, sconstruct_path, local_sconstruct_path
                    )
                else:
                    click.secho("Abort!", fg="red")

        else:
            self._copy_sconstruct_file(
                sconstruct_name, sconstruct_path, local_sconstruct_path
            )

    def create_ini(self, board, project_dir="", sayyes=False):
        """Creates a new apio project file"""

        project_dir = util.check_dir(project_dir)

        ini_path = util.safe_join(project_dir, PROJECT_FILENAME)

        # Check board
        boards = Resources().boards
        if board not in boards.keys():
            click.secho("Error: no such board '{}'".format(board), fg="red")
            sys.exit(1)

        if isfile(ini_path):
            # -- If sayyes, skip the question
            if sayyes:
                self._create_ini_file(board, ini_path, PROJECT_FILENAME)
            else:
                click.secho(
                    "Warning: {} file already exists".format(PROJECT_FILENAME),
                    fg="yellow",
                )
                if click.confirm("Do you want to replace it?"):
                    self._create_ini_file(board, ini_path, PROJECT_FILENAME)
                else:
                    click.secho("Abort!", fg="red")
        else:
            self._create_ini_file(board, ini_path, PROJECT_FILENAME)

    @staticmethod
    def _create_ini_file(board, ini_path, ini_name):
        click.secho("Creating {} file ...".format(ini_name))
        with open(ini_path, "w") as file:
            config = ConfigParser.ConfigParser()
            config.add_section("env")
            config.set("env", "board", board)
            config.write(file)
            click.secho(
                "File '{}' has been successfully created!".format(ini_name),
                fg="green",
            )

    @staticmethod
    def _copy_sconstruct_file(
        sconstruct_name, sconstruct_path, local_sconstruct_path
    ):
        click.secho("Creating {} file ...".format(sconstruct_name))
        with open(sconstruct_path, "w") as sconstruct:
            with open(local_sconstruct_path, "r") as local_sconstruct:
                sconstruct.write(local_sconstruct.read())
                click.secho(
                    "File '{}' has been successfully created!".format(
                        sconstruct_name
                    ),
                    fg="green",
                )

    def read(self):
        """Read the project config file"""

        # -- If no project finel found, just return
        if not isfile(PROJECT_FILENAME):
            print("Info: No {} file".format(PROJECT_FILENAME))
            return

        # -- Read stored board
        board = self._read_board()

        # -- Update board
        self.board = board
        if not board:
            print("Error: invalid {} project file".format(PROJECT_FILENAME))
            print("No 'board' field defined in project file")
            sys.exit(1)

    @staticmethod
    def _read_board():
        board = ""

        # -- Read config file: old JSON format
        with open(PROJECT_FILENAME, "r") as file:
            try:
                data = json.loads(file.read())
                board = data.get("board")
            except Exception:
                pass

        # -- Read config file: new CFG format
        if board == "":
            try:
                config = ConfigParser.ConfigParser()
                config.read(PROJECT_FILENAME)
                board = config.get("env", "board")
            except Exception:
                print(
                    "Error: invalid {} project file".format(PROJECT_FILENAME)
                )
                sys.exit(1)

        return board



import pytest

from os import getcwd, listdir, mkdir
from os.path import join, isfile, isdir, getsize

import click

from apio.commands.install import cli as cmd_install
from apio.commands.uninstall import cli as cmd_uninstall
from apio.commands.init import cli as cmd_init
from apio.commands.upload import cli as cmd_upload
from apio.commands.examples import cli as cmd_examples


def validate_files_leds(apioproject_dir):
    path = join(apioproject_dir, 'leds.v')
    assert isfile(path) and getsize(path) > 0


def validate_dir_leds(apioproject_dir):
    path = join(apioproject_dir, 'leds')
    assert isdir(path) and len(listdir(path)) > 0


def test_complete(clirunner, validate_cliresult, configenv, offline):
    if offline:
        pytest.skip('requires internet connection')

    with clirunner.isolated_filesystem():
        configenv()

        # apio uninstall examples
        result = clirunner.invoke(
            cmd_uninstall, ['examples'], input='y')
        assert 'Do you want to continue?' in result.output
        assert 'Error: package \'examples\' is not installed' in result.output

        # apio install examples@X
        result = clirunner.invoke(cmd_install, ['examples@X'])
        assert 'Warning: package \'examples\' version X' in result.output
        assert 'does not match the semantic version' in result.output

        # apio install examples@0.0.7
        result = clirunner.invoke(cmd_install, ['examples@0.0.7'])
        validate_cliresult(result)
        assert 'Installing examples package' in result.output
        if click.__version__ != '7.0':
            assert 'Downloading' in result.output
            assert 'Unpacking' in result.output
        assert 'has been successfully installed!' in result.output

        # apio install examples
        result = clirunner.invoke(cmd_install, ['examples'])
        validate_cliresult(result)
        assert 'Installing examples package' in result.output
        if click.__version__ != '7.0':
            assert 'Downloading' in result.output
            assert 'Unpacking' in result.output
        assert 'has been successfully installed!' in result.output

        # apio install examples
        result = clirunner.invoke(cmd_install, ['examples'])
        validate_cliresult(result)
        assert 'Installing examples package' in result.output
        assert 'Already installed. Version ' in result.output

        # apio install examples -p windows
        result = clirunner.invoke(cmd_install, [
            'examples', '--platform', 'windows', '--force'])
        validate_cliresult(result)
        assert 'Installing examples package' in result.output
        if click.__version__ != '7.0':
            assert 'Downloading' in result.output
            assert 'Unpacking' in result.output
        assert 'has been successfully installed!' in result.output

        # apio install --list
        result = clirunner.invoke(cmd_install, ['--list'])
        validate_cliresult(result)

        # apio init --board icezum
        result = clirunner.invoke(cmd_init, ['--board', 'icezum'])
        validate_cliresult(result)
        assert 'Creating apio.ini file ...' in result.output
        assert 'has been successfully created!' in result.output

        # apio upload
        result = clirunner.invoke(cmd_upload)
        assert result.exit_code == 1

        # apio examples --list
        result = clirunner.invoke(cmd_examples, ['--list'])
        validate_cliresult(result)
        assert 'leds' in result.output
        assert 'icezum' in result.output

        # apio examples --files missing_example
        result = clirunner.invoke(cmd_examples, ['--files', 'missing_example'])
        validate_cliresult(result)
        assert 'Warning: this example does not exist' in result.output

        # apio examples --files leds
        result = clirunner.invoke(cmd_examples, ['--files', 'leds'])
        validate_cliresult(result)
        assert 'Copying leds example files ...' in result.output
        assert 'have been successfully created!' in result.output
        validate_files_leds(getcwd())

        # apio examples --dir leds
        result = clirunner.invoke(cmd_examples, ['--dir', 'leds'])
        validate_cliresult(result)
        assert 'Creating leds directory ...' in result.output
        assert 'has been successfully created!' in result.output
        validate_dir_leds(getcwd())

        # apio examples --dir leds
        result = clirunner.invoke(cmd_examples, ['--dir', 'leds'], input='y')
        validate_cliresult(result)
        assert 'Warning: leds directory already exists' in result.output
        assert 'Do you want to replace it?' in result.output
        assert 'Creating leds directory ...' in result.output
        assert 'has been successfully created!' in result.output
        validate_dir_leds(getcwd())

        dir_name = 'tmp'
        mkdir(dir_name)

        # apio examples --files leds --project-dir=tmp
        result = clirunner.invoke(
            cmd_examples, ['--files', 'leds', '--project-dir=tmp'])
        validate_cliresult(result)
        assert 'Copying leds example files ...' in result.output
        assert 'have been successfully created!' in result.output
        validate_files_leds(join(getcwd(), dir_name))

        # apio examples --dir leds --project-dir=tmp
        result = clirunner.invoke(
            cmd_examples, ['--dir', 'leds', '--project-dir=tmp'])
        validate_cliresult(result)
        assert 'Creating leds directory ...' in result.output
        assert 'has been successfully created!' in result.output
        validate_dir_leds(join(getcwd(), dir_name))

        # apio uninstall examples
        result = clirunner.invoke(cmd_uninstall, ['examples'], input='n')
        validate_cliresult(result)
        assert 'Abort!' in result.output

        # apio uninstall examples
        result = clirunner.invoke(cmd_uninstall, ['examples'], input='y')
        validate_cliresult(result)
        assert 'Uninstalling examples package' in result.output
        assert 'Do you want to continue?' in result.output
        assert 'has been successfully uninstalled!' in result.output



from apio.commands.lint import cli as cmd_lint


def test_lint(clirunner, configenv):
    with clirunner.isolated_filesystem():
        configenv()
        result = clirunner.invoke(cmd_lint, ['--board', 'icezum'])
        assert result.exit_code != 0
        if result.exit_code == 1:
            assert 'apio install verilator' in result.output



# -*- coding: utf-8 -*-
# -- This file is part of the Apio project
# -- (C) 2016-2019 FPGAwars
# -- Author Jesús Arroyo
# -- Licence GPLv2

import click

from apio.profile import Profile


@click.command("config")
@click.pass_context
@click.option(
    "-l", "--list", is_flag=True, help="List all configuration parameters."
)
@click.option(
    "-v",
    "--verbose",
    type=click.Choice(["0", "1"]),
    help="Verbose mode: `0` General, `1` Information.",
)
@click.option(
    "-e",
    "--exe",
    type=click.Choice(["default", "native"]),
    help="Configure executables: `default` selects apio packages, "
    + "`native` selects system binaries.",
)
def cli(ctx, list, verbose, exe):
    """Apio configuration."""

    if list:  # pragma: no cover
        profile = Profile()
        profile.list()
    elif verbose:  # pragma: no cover
        profile = Profile()
        profile.add_config("verbose", verbose)
    elif exe:  # pragma: no cover
        profile = Profile()
        profile.add_config("exe", exe)
    else:
        click.secho(ctx.get_help())



from apio.commands.uninstall import cli as cmd_uninstall


def test_uninstall(clirunner, validate_cliresult):
    result = clirunner.invoke(cmd_uninstall)
    validate_cliresult(result)


def test_uninstall_list(clirunner, validate_cliresult, configenv):
    with clirunner.isolated_filesystem():
        configenv()
        result = clirunner.invoke(cmd_uninstall, ['--list'])
        validate_cliresult(result)


def test_uninstall_wrong_package(clirunner, configenv):
    with clirunner.isolated_filesystem():
        configenv()
        result = clirunner.invoke(
            cmd_uninstall, ['missing_package'], input='y')
        assert result.exit_code == 1
        assert 'Do you want to continue?' in result.output
        assert 'Error: no such package' in result.output
